import json
from pathlib import Path
from typing import Any, Dict, List, Optional, Text

from neuralspace.constants import TYPE
from neuralspace.datamodels import DatasetType


class DataConverter:
    LOOKUP_FILE = "lookup.json"
    REGEX_FILE = "regex.json"
    SYNONYM_FILE = "synonym.json"
    NLU_FILE = "nlu.json"

    @staticmethod
    def __training_data_converter(final_data) -> List[Dict[Text, Any]]:
        NotImplementedError("Training data converter is not implemented")
        pass

    def __regex_converter(self, final_data: Dict[Text, Any]) -> List[Dict[Text, Any]]:
        NotImplementedError("Regex converter is not implemented")

    def __synonym_converter(self, final_data: Dict[Text, Any]) -> List[Dict[Text, Any]]:
        NotImplementedError("synonym converter is not implemented")

    def __lookup_converter(self, final_data: Dict[Text, Any]) -> List[Dict[Text, Any]]:
        NotImplementedError("lookup converter is not implemented")

    def convert(self, input_path: Path, output_path: Path, dataset_type: DatasetType):
        NotImplementedError("lookup converter is not implemented")

    def convert_multiple(
        self, input_path: List[Path], output_path: Path, dataset_type: DatasetType
    ):
        NotImplementedError("lookup converter is not implemented")

    @staticmethod
    def count(dataset):
        return len(dataset)

    @staticmethod
    def save_converted_data(
        output_directory: Path,
        lookup_data: Optional[List[Dict[Text, Any]]] = None,
        regex_data: Optional[List[Dict[Text, Any]]] = None,
        synonym_data: Optional[List[Dict[Text, Any]]] = None,
        nlu_data: Optional[List[Dict[Text, Any]]] = None,
    ):
        output_directory.mkdir(parents=True, exist_ok=True)

        if lookup_data is not None:
            with open(output_directory / DataConverter.LOOKUP_FILE, "w") as f:
                json.dump(lookup_data, f, indent=4, ensure_ascii=False)
        if regex_data is not None:
            with open(output_directory / DataConverter.REGEX_FILE, "w") as f:
                json.dump(regex_data, f, indent=4, ensure_ascii=False)
        if synonym_data is not None:
            with open(output_directory / DataConverter.SYNONYM_FILE, "w") as f:
                json.dump(synonym_data, f, indent=4, ensure_ascii=False)
        if nlu_data is not None:
            with open(output_directory / DataConverter.NLU_FILE, "w") as f:
                json.dump(nlu_data, f, indent=4, ensure_ascii=False)

    def set_data_type(self, dataset_type: DatasetType = "train"):
        for example in self.nlu_data:
            example[TYPE] = dataset_type



LANGUAGE_DETECTION_APP_URL = "api/language-detection/v1/detect"
LANGUAGE_DETECTION_LANGUAGE_CATALOG_URL = (
    "api/language-detection/v1/supported-languages"
)
LANG_DETECTION_TEXT_COMMAND = "neuralspace language-detection parse --help"

SUPPORTED_LANGUAGES = [
    "af",
    "als",
    "am",
    "an",
    "ar",
    "arz",
    "as",
    "ast",
    "av",
    "az",
    "azb",
    "ba",
    "bar",
    "bcl",
    "be",
    "bg",
    "bh",
    "bn",
    "bo",
    "bpy",
    "br",
    "bs",
    "bxr",
    "ca",
    "cbk",
    "ce",
    "ceb",
    "ckb",
    "co",
    "cs",
    "cv",
    "cy",
    "da",
    "de",
    "diq",
    "dsb",
    "dty",
    "dv",
    "el",
    "eml",
    "en",
    "eo",
    "es",
    "et",
    "eu",
    "fa",
    "fi",
    "fr",
    "frr",
    "fy",
    "ga",
    "gd",
    "gl",
    "gn",
    "gom",
    "gu",
    "gv",
    "he",
    "hi",
    "hif",
    "hr",
    "hsb",
    "ht",
    "hu",
    "hy",
    "ia",
    "id",
    "ie",
    "ilo",
    "io",
    "is",
    "it",
    "ja",
    "jbo",
    "jv",
    "ka",
    "kk",
    "km",
    "kn",
    "ko",
    "krc",
    "ku",
    "kv",
    "kw",
    "ky",
    "la",
    "lb",
    "lez",
    "li",
    "lmo",
    "lo",
    "lrc",
    "lt",
    "lv",
    "mai",
    "mg",
    "mhr",
    "min",
    "mk",
    "ml",
    "mn",
    "mr",
    "mrj",
    "ms",
    "mt",
    "mwl",
    "my",
    "myv",
    "mzn",
    "nah",
    "nap",
    "nds",
    "ne",
    "new",
    "nl",
    "nn",
    "no",
    "oc",
    "or",
    "os",
    "pa",
    "pam",
    "pfl",
    "pl",
    "pms",
    "pnb",
    "ps",
    "pt",
    "qu",
    "rm",
    "ro",
    "ru",
    "rue",
    "sa",
    "sah",
    "sc",
    "scn",
    "sco",
    "sd",
    "si",
    "sk",
    "sl",
    "so",
    "sq",
    "sr",
    "su",
    "sv",
    "sw",
    "ta",
    "te",
    "tg",
    "th",
    "tk",
    "tl",
    "tr",
    "tt",
    "tyv",
    "ug",
    "uk",
    "ur",
    "uz",
    "vec",
    "vep",
    "vi",
    "vls",
    "vo",
    "wa",
    "war",
    "wuu",
    "xal",
    "xmf",
    "yi",
    "yo",
    "yue",
    "zh",
]



import json
import time as t
from copy import copy
from pathlib import Path
from typing import Any, Dict, List, Text, Tuple

import numpy
import requests
from rich.console import Console
from rich.table import Table

from neuralspace.constants import (
    AUTHORIZATION,
    BAG,
    COMMON_HEADERS,
    COUNT,
    CROSS,
    DATA,
    ENTITIES,
    ENTITY,
    ENTITY_TYPE,
    EXAMPLES,
    FILTER,
    GREEN_TICK,
    INFO,
    LANGUAGE,
    LIST_ENTITY_URL,
    LOOKUP,
    PAGE_NUMBER,
    PAGE_SIZE,
    PRE_TRAINED,
    REGEX,
    SEARCH,
    neuralspace_url,
)
from neuralspace.utils import get_auth_token, is_success_status

console = Console()


def convert_files(convert, rasa_data):
    lookup_data = convert.lookup_converter(rasa_data.lookup_tables)
    regex_data = convert.regex_converter(rasa_data.regex_features)
    synonym_data = convert.synonym_converter(rasa_data.entity_synonyms)
    training_data = convert.training_data_converter(rasa_data)
    return lookup_data, regex_data, synonym_data, training_data


def __replace_entities_with_mapping(
    entities: List[Dict[Text, Any]], entity_mapping: Dict[Text, Text]
):
    for entity_value in entities:
        if entity_value[ENTITY] in entity_mapping:
            entity_value[ENTITY] = entity_mapping[entity_value[ENTITY]]


def map_entities(nlu_data: List[Dict[Text, Any]], entity_mapping: Dict[Text, Text]):
    for example in nlu_data:
        entities = example.get(ENTITIES, [])
        if entities:
            __replace_entities_with_mapping(entities, entity_mapping)
    return nlu_data


def get_entity_list_from_remote(
    entity_type: Text,
    language: Text,
    keyword_search: Text = "",
    page_size: int = 1,
    page_number: int = 10,
) -> Tuple[List[Dict[Text, Any]], int]:
    payload = {
        FILTER: {ENTITY_TYPE: entity_type, LANGUAGE: language},
        SEARCH: keyword_search,
        PAGE_NUMBER: page_number,
        PAGE_SIZE: page_size,
    }
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()

    response = requests.request(
        "POST",
        f"{neuralspace_url()}/{LIST_ENTITY_URL}",
        headers=HEADERS,
        data=json.dumps(payload, ensure_ascii=False),
    )
    json_response = response.json(encoding="utf-8")

    if not is_success_status(response.status_code):
        raise Exception(
            f"Error while fetching entities. "
            f"Error message: {json.dumps(json_response, indent=4)}"
        )
    return json_response[DATA][ENTITIES], json_response[DATA][COUNT]


def auto_tag_pretrained_entities(data: List[Dict[Text, Any]], language: Text):
    sample_pretrained_entities, count = get_entity_list_from_remote(
        entity_type=PRE_TRAINED, language=language
    )
    all_pretrained_entities, _ = get_entity_list_from_remote(
        entity_type=PRE_TRAINED,
        language=language,
        page_number=1,
        page_size=count,
    )
    all_pretrained_entity_names = {
        entity[ENTITY]: entity[ENTITY] for entity in all_pretrained_entities
    }
    for example in data:
        entities = example.get(ENTITIES, [])
        for entity_value in entities:
            if entity_value[ENTITY] in all_pretrained_entity_names:
                entity_value[ENTITY] = all_pretrained_entity_names[entity_value[ENTITY]]
                entity_value[ENTITY_TYPE] = PRE_TRAINED
    return data


def auto_tag_lookup(
    data: List[Dict[Text, Any]], lookup_entities: List[Dict[Text, Any]]
):
    all_lookup_entity_names = {
        entity[ENTITY]: entity[ENTITY] for entity in lookup_entities
    }
    for example in data:
        entities = example.get(ENTITIES, [])
        for entity_value in entities:
            if (
                entity_value[ENTITY] in all_lookup_entity_names
                and ENTITY_TYPE not in entity_value
            ):
                entity_value[ENTITY] = all_lookup_entity_names[entity_value[ENTITY]]
                entity_value[ENTITY_TYPE] = LOOKUP
    return data


def auto_tag_regex(data: List[Dict[Text, Any]], regex_entities: List[Dict[Text, Any]]):
    all_regex_entity_names = {
        entity[ENTITY]: entity[ENTITY] for entity in regex_entities
    }
    for example in data:
        entities = example.get(ENTITIES, [])
        for entity_value in entities:
            if (
                entity_value[ENTITY] in all_regex_entity_names
                and ENTITY_TYPE not in entity_value
            ):
                entity_value[ENTITY] = all_regex_entity_names[entity_value[ENTITY]]
                entity_value[ENTITY_TYPE] = REGEX
    return data


def remove_nlu_duplicates(
    data_to_clear_duplicates: List[Dict[Text, Any]]
) -> List[Dict[Text, Any]]:
    numpy_format_for_the_list: numpy.array = numpy.array(data_to_clear_duplicates)
    for data in data_to_clear_duplicates:
        indexes_where_the_data_present = numpy.where(data == numpy_format_for_the_list)
        for index in indexes_where_the_data_present[0][1:]:
            numpy_format_for_the_list = numpy.delete(numpy_format_for_the_list, index)
    return list(numpy_format_for_the_list)


def remove_entity_duplicates(
    entities: List[Dict[Text, Any]],
) -> List[Dict[Text, Any]]:
    for entity in entities:
        entity[EXAMPLES] = sorted(list(set(entity[EXAMPLES])), key=str.casefold)
    return entities


def print_converted_data_inference(from_platform: Text, output_path: Path, converter):
    console.print(
        f"> {GREEN_TICK} SUCCESS Your {from_platform} dataset is now converted to Neuralspace-platform data"
    )
    console.print(f"> {INFO} Converted dataset is {BAG} stored in: {output_path}")
    table = Table(show_header=True, header_style="bold #c47900", show_lines=True)
    table.add_column("Contents")
    table.add_column("Total")
    table.add_row(
        "Number of examples before removing duplicates",
        str(converter.number_of_nlu_examples_before_duplicate_check),
    )
    table.add_row(
        "Number of examples after removing the duplicates",
        str(converter.number_of_nlu_examples_after_duplicate_check),
    )
    table.add_row("Number of intent present", str(converter.number_of_unique_intent()))
    console.print(table)
    console.print(
        "Number of duplicates found: ",
        str(
            converter.number_of_nlu_examples_before_duplicate_check
            - converter.number_of_nlu_examples_after_duplicate_check
        ),
    )
    console.print("\nChecking Files... \n")
    console.print("> Lookups", end=" ")
    t.sleep(1)
    console.print(f"{GREEN_TICK} ") if converter.lookup_data else console.print(
        f"{CROSS} "
    )
    console.print("> Synonyms", end=" ")
    t.sleep(1)
    console.print(f"{GREEN_TICK} ") if converter.synonym_data else console.print(
        f"{CROSS} "
    )
    console.print("> Regex", end=" ")
    t.sleep(1)
    console.print(f"{GREEN_TICK} ") if converter.regex_data else console.print(
        f"{CROSS} "
    )


def print_model_ids(model_ids: List[Text]):
    table = Table(show_header=True, header_style="bold #c47900", show_lines=True)
    table.add_column("Index")
    table.add_column("Model Id")
    for index, single_model_id in enumerate(model_ids):
        table.add_row(str(index), single_model_id)
    console.print(table)



import asyncio
from typing import Text

import click

from neuralspace.apis import get_async_http_session
from neuralspace.translation.apis import get_languages, get_translation_response
from neuralspace.translation.constants import SUPPORTED_LANGUAGES
from neuralspace.utils import setup_logger


@click.group(name="translation")
def translation():
    pass


@translation.command(
    name="parse",
)
@click.option(
    "-t",
    "--text",
    type=click.STRING,
    required=True,
    help="text to be translated",
)
@click.option(
    "-src",
    "--src-language",
    type=click.STRING,
    required=True,
    help="language that the text is in",
)
@click.option(
    "-tgt",
    "--tgt-language",
    type=click.STRING,
    required=True,
    help="language you wish to translate to",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def translate(text: Text, src_language: Text, tgt_language: Text, log_level: Text):
    setup_logger(log_level=log_level)
    if not src_language:
        ValueError(
            f"Source Language is mandatory. You can select from {SUPPORTED_LANGUAGES}"
        )
    if not tgt_language:
        ValueError(
            f"Target Language is mandatory. You can select from {SUPPORTED_LANGUAGES}"
        )
    if not text:
        ValueError("Please enter the text to translate")
    loop = asyncio.new_event_loop()
    loop.run_until_complete(get_translation_response(text, src_language, tgt_language))
    loop.run_until_complete(get_async_http_session().close())


@translation.command(name="get-languages")
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def translation_get_languages(log_level: Text):
    setup_logger(log_level=log_level)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(get_languages())
    loop.run_until_complete(get_async_http_session().close())



from dataclasses import dataclass
from pathlib import Path
from typing import Dict, List, Text

import datasets
from datasets import load_dataset, load_dataset_builder

from neuralspace.constants import NLU_FILE, TEST, TRAIN, neuralspace_home
from neuralspace.nlu.converters.csvconverter import CsvConverter


@dataclass
class HuggingfaceDatasetCollection:
    subset: Text
    ner_dict: Dict
    train_dataset: datasets.arrow_dataset.Dataset
    test_dataset: datasets.arrow_dataset.Dataset


class HuggingfaceNerConverter:
    INTERMEDIATE_CSV_FOLDER = neuralspace_home() / "csv_dataset"

    def __init__(self, language: Text, max_entities_in_text: int):
        self.language = language
        self.max_entities_in_text = max_entities_in_text

    def ner_list_to_dict(self, ner_list: List) -> Dict:
        ner_dict = {}
        for index, entity in enumerate(ner_list):
            ner_dict[index] = entity
        return ner_dict

    def load_huggingface_dataset(
        self,
        dataset_name: Text,
        subsets: List,
        num_train_examples: int,
        num_test_examples: int,
    ) -> HuggingfaceDatasetCollection:
        hf_dataset = []
        for subset in subsets:
            subset_ner_tags = (
                load_dataset_builder(dataset_name, subset)
                .info.features["ner_tags"]
                .feature.names
            )
            dataset = load_dataset(dataset_name, subset)
            if num_train_examples == -1:
                train_dataset = dataset[TRAIN]
            else:
                train_dataset = dataset[TRAIN][:num_train_examples]

            if TEST in dataset.keys():
                if num_test_examples == -1:
                    test_dataset = dataset[TEST]
                else:
                    test_dataset = dataset[TEST][:num_test_examples]
            else:
                test_dataset = None

            ner_dict = self.ner_list_to_dict(subset_ner_tags)
            hf_dataset.append(
                HuggingfaceDatasetCollection(
                    subset=subset,
                    ner_dict=ner_dict,
                    train_dataset=train_dataset,
                    test_dataset=test_dataset,
                )
            )
        return hf_dataset

    def __convert_tokens_and_ner_tags_to_text(
        self, tokens: List, ner_tags: List, ner_dict: Dict
    ):
        text = ""
        for token, ner_tag in zip(tokens, ner_tags):
            if ner_dict[ner_tag] != "O":
                text += f"[{token}]({ner_dict[ner_tag]})" + " "
            elif token == ".":
                text = text[:-1] + token
            else:
                text += token + " "
        return text

    def convert_huggingface_dataset_to_list(
        self, hf_dataset: HuggingfaceDatasetCollection
    ):
        train_list = []
        test_list = []
        for subset_dataset in hf_dataset:
            ner_dict = subset_dataset.ner_dict
            train_dataset = subset_dataset.train_dataset
            test_dataset = subset_dataset.test_dataset

            train_tokens_list = train_dataset["tokens"]
            train_ner_tags_list = train_dataset["ner_tags"]

            for tokens, ner_tags in zip(train_tokens_list, train_ner_tags_list):
                text = self.__convert_tokens_and_ner_tags_to_text(
                    tokens, ner_tags, ner_dict
                )
                train_list.append(text)

            if test_dataset is not None:
                test_tokens_list = test_dataset["tokens"]
                test_ner_tags_list = test_dataset["ner_tags"]
                for tokens, ner_tags in zip(test_tokens_list, test_ner_tags_list):
                    text = self.__convert_tokens_and_ner_tags_to_text(
                        tokens, ner_tags, ner_dict
                    )
                    test_list.append(text)

        train_list = list(set(train_list))
        test_list = list(set(test_list))

        return train_list, test_list

    def __preprocess_text(self, text):
        text = text.replace(" ,", "")
        text = text.replace(",", "")
        text = text.replace('"', "").replace("'", "")
        return text

    def write_list_to_ns_format_csv(self, data: List, output_file_path: Path):
        with open(str(output_file_path), "w") as f:
            f.write(f"label,text\n")  # noqa
            for i, text in enumerate(data):
                text = self.__preprocess_text(text)
                if text.rstrip() != "" and text.count("[") <= self.max_entities_in_text:
                    if i % 2 == 0:
                        f.write(f"intent_1,{text}\n")
                    else:
                        f.write(f"intent_2,{text}\n")

    def convert_csv_to_ns_format(
        self, csv_folder_path: Path, output_folder_path, test_data_present: bool = True
    ):
        converter = CsvConverter(
            language=self.language,
            entity_mapping=False,
            auto_tag_entities=False,
            ignore_missing_examples=False,
            ignore_swapped_columns=False,
        )
        converter.convert(
            input_path=Path(csv_folder_path / TRAIN),
            output_directory=output_folder_path / TRAIN,
            dataset_type=TRAIN,
        )
        if test_data_present:
            converter.convert(
                input_path=Path(csv_folder_path / TEST),
                output_directory=output_folder_path / TEST,
                dataset_type=TEST,
            )

    def convert(
        self,
        hf_dataset_name: Text,
        subsets: Text,
        num_train_examples: int,
        num_test_examples: int,
        output_directory: Path,
    ):
        hf_dataset = self.load_huggingface_dataset(
            hf_dataset_name, subsets, num_train_examples, num_test_examples
        )
        train_list, test_list = self.convert_huggingface_dataset_to_list(hf_dataset)
        tmp_train_dataset_path = HuggingfaceNerConverter.INTERMEDIATE_CSV_FOLDER / TRAIN
        tmp_test_dataset_path = HuggingfaceNerConverter.INTERMEDIATE_CSV_FOLDER / TEST

        Path(tmp_train_dataset_path).mkdir(parents=True, exist_ok=True)
        Path(tmp_test_dataset_path).mkdir(parents=True, exist_ok=True)

        self.write_list_to_ns_format_csv(
            train_list, Path(tmp_train_dataset_path) / NLU_FILE
        )
        self.write_list_to_ns_format_csv(
            test_list, Path(tmp_test_dataset_path) / NLU_FILE
        )
        if test_list == []:
            test_data_present = False
        else:
            test_data_present = True

        self.convert_csv_to_ns_format(
            HuggingfaceNerConverter.INTERMEDIATE_CSV_FOLDER,
            output_directory,
            test_data_present,
        )



from pathlib import Path
from typing import Any, Dict, List, Optional, Text

from rasa.shared.nlu.training_data.training_data import TrainingData

from neuralspace.constants import (
    ELEMENTS,
    ENTITIES,
    ENTITY,
    ENTITY_TYPE,
    EXAMPLES,
    INTENT,
    LANGUAGE,
    LOOKUP,
    NAME,
    PATTERN,
    REGEX,
    SYNONYM,
)
from neuralspace.datamodels import DatasetType
from neuralspace.nlu.converters.base import DataConverter
from neuralspace.nlu.utils import (
    auto_tag_lookup,
    auto_tag_pretrained_entities,
    auto_tag_regex,
    map_entities,
    remove_entity_duplicates,
    remove_nlu_duplicates,
)


class RasaConverter(DataConverter):
    def __init__(
        self,
        language: Text,
        entity_mapping: Optional[Dict[Text, Text]],
        auto_tag_entities: bool,
    ):
        self.nlu_data = []
        self.lookup_data = []
        self.regex_data = []
        self.synonym_data = []
        self.language = language
        self.entity_mapping = entity_mapping
        self.auto_tag_entities = auto_tag_entities
        self.number_of_nlu_examples_before_duplicate_check = 0
        self.number_of_nlu_examples_after_duplicate_check = 0

    @property
    def all_entity_types(self) -> List[Text]:
        all_entities = []
        all_entities += [entity[ENTITY] for entity in self.lookup_data]
        all_entities += [entity[ENTITY] for entity in self.regex_data]
        all_entities += [entity[ENTITY] for entity in self.synonym_data]
        all_nlu_entities = []
        for example in self.nlu_data:
            if ENTITIES in example:
                for entity in example[ENTITIES]:
                    all_nlu_entities.append(entity[ENTITY])
        all_entities += list(set(all_nlu_entities))
        return list(set(all_nlu_entities))

    def __lookup_converter(
        self, entities: List[Dict[Text, Any]]
    ) -> List[Dict[Text, Any]]:
        formatted_lookup_entities = []
        for lookup_entity in entities:
            entity = {
                ENTITY: lookup_entity[NAME],
                EXAMPLES: lookup_entity[ELEMENTS],
                LANGUAGE: self.language,
                ENTITY_TYPE: LOOKUP,
            }
            formatted_lookup_entities.append(entity)
        return formatted_lookup_entities

    def __regex_converter(
        self, entities: List[Dict[Text, Any]]
    ) -> List[Dict[Text, Any]]:
        """
        Rasa's format -> [{‘name’: ‘help’, ‘pattern’: ‘\\bhelp\\b’}, {‘name’: ‘help’, ‘pattern’: ‘\\df’}]
        NeuralSpace format ->
        {
            "entity": "entity-name",
            "examples": [],
            "language": "language-code",
            "entityType": "regex"
        }
        """
        all_regexes = {}
        for regex_entity in entities:
            if regex_entity[NAME] not in all_regexes:
                all_regexes[regex_entity[NAME]] = {
                    ENTITY: regex_entity[NAME],
                    EXAMPLES: [regex_entity[PATTERN]],
                    LANGUAGE: self.language,
                    ENTITY_TYPE: REGEX,
                }
            else:
                all_regexes[regex_entity[NAME]][EXAMPLES].append(regex_entity[PATTERN])
        return list(all_regexes.values())

    def __synonym_converter(self, synonyms: Dict[Text, Any]) -> List[Dict[Text, Any]]:
        """
        Rasa's format -> {'super': 'good', 'great': 'good', 'done good': 'good'}
        Note that in rasa's format the value in teh dictionary is the actual name of the synonym entity

        NeuralSpace format ->
        {
            "entity": "synonym-name",
            "examples": [],
            "language": "language-code",
            "entityType": "synonyms"
        }
        """
        formatted_synonyms = {}
        for synonym, entity_name in synonyms.items():
            if entity_name not in formatted_synonyms:
                formatted_synonyms[entity_name] = {
                    ENTITY: entity_name,
                    EXAMPLES: [synonym],
                    LANGUAGE: self.language,
                    ENTITY_TYPE: SYNONYM,
                }
            else:
                formatted_synonyms[entity_name][EXAMPLES].append(synonym)
        return list(formatted_synonyms.values())

    @staticmethod
    def __nlu_data_converter(
        dataset: TrainingData,
    ):
        formatted_nlu_data = []
        for example in dataset.nlu_examples:
            formatted_nlu_data.append(example.as_dict())
        return formatted_nlu_data

    def number_of_unique_intent(self):
        return len(set([example[INTENT] for example in self.nlu_data]))

    def post_process(self, dataset_type, output_directory):
        if self.entity_mapping:
            self.nlu_data = map_entities(self.nlu_data, self.entity_mapping)
        if self.auto_tag_entities:
            self.nlu_data = auto_tag_pretrained_entities(self.nlu_data, self.language)
            self.nlu_data = auto_tag_lookup(self.nlu_data, self.lookup_data)
            self.nlu_data = auto_tag_regex(self.nlu_data, self.regex_data)
        self.number_of_nlu_examples_before_duplicate_check = self.count(self.nlu_data)
        self.nlu_data = remove_nlu_duplicates(self.nlu_data)
        self.number_of_nlu_examples_after_duplicate_check = self.count(self.nlu_data)
        self.lookup_data = remove_entity_duplicates(
            remove_nlu_duplicates(self.lookup_data)
        )
        self.regex_data = remove_entity_duplicates(
            remove_nlu_duplicates(self.regex_data)
        )
        self.synonym_data = remove_entity_duplicates(
            remove_nlu_duplicates(self.synonym_data)
        )
        self.set_data_type(dataset_type)
        self.save_converted_data(
            output_directory=output_directory,
            nlu_data=self.nlu_data,
            lookup_data=self.lookup_data,
            synonym_data=self.synonym_data,
            regex_data=self.regex_data,
        )

    def convert(
        self, input_path: List[Path], output_directory: Path, dataset_type: DatasetType
    ):
        from rasa.shared.importers.utils import training_data_from_paths

        for single_path in input_path:
            data: TrainingData = training_data_from_paths(
                [str(single_path)], language=self.language
            )
            self.nlu_data += self.__nlu_data_converter(data)
            self.lookup_data += self.__lookup_converter(data.lookup_tables)
            self.regex_data += self.__regex_converter(data.regex_features)
            self.synonym_data += self.__synonym_converter(data.entity_synonyms)

        self.post_process(dataset_type=dataset_type, output_directory=output_directory)

    def component_convert(self, training_data: TrainingData, dataset_type: DatasetType):
        self.nlu_data += self.__nlu_data_converter(training_data)
        self.lookup_data += self.__lookup_converter(training_data.lookup_tables)
        self.regex_data += self.__regex_converter(training_data.regex_features)
        self.synonym_data += self.__synonym_converter(training_data.entity_synonyms)

        if self.entity_mapping:
            self.nlu_data = map_entities(self.nlu_data, self.entity_mapping)

        if self.auto_tag_entities:
            self.nlu_data = auto_tag_pretrained_entities(self.nlu_data, self.language)
            self.nlu_data = auto_tag_lookup(self.nlu_data, self.lookup_data)
            self.nlu_data = auto_tag_regex(self.nlu_data, self.regex_data)

        self.set_data_type(dataset_type)
        self.nlu_data = remove_nlu_duplicates(self.nlu_data)
        self.lookup_data = remove_entity_duplicates(
            remove_nlu_duplicates(self.lookup_data)
        )
        self.regex_data = remove_entity_duplicates(
            remove_nlu_duplicates(self.regex_data)
        )
        self.synonym_data = remove_entity_duplicates(
            remove_nlu_duplicates(self.synonym_data)
        )



import asyncio
import json
import typing
from contextvars import Context
from inspect import Parameter
from pathlib import Path
from typing import List, Text

import click
import randomname
from rich.console import Console

from neuralspace.apis import get_async_http_session
from neuralspace.constants import (
    ERROR,
    NAME_GENERATOR_CONFIG,
    NER_TYPES,
    ORANGE3_END,
    ORANGE3_START,
    RED_END,
    RED_START,
)
from neuralspace.datamodels import DatasetTypes
from neuralspace.nlu.apis import (
    create_project,
    delete_examples,
    delete_models,
    delete_project,
    deploy,
    get_languages,
    list_examples,
    list_models,
    list_projects,
    parse,
    train_model,
    upload_dataset,
    wait_till_training_completes,
)
from neuralspace.nlu.constants import SUPPORTED_LANGUAGES, TRAINING_PROGRESS
from neuralspace.nlu.utils import print_converted_data_inference
from neuralspace.utils import get_entity_list, print_logo, setup_logger

console = Console()


class MappingClass(click.ParamType):
    name: Text = "mapping_class"

    def convert(
        self,
        value: typing.Any,
        param: typing.Optional["Parameter"],
        ctx: typing.Optional["Context"],
    ) -> typing.Union[typing.Dict[Text, Text], bool]:
        entity_mapping = value.split(",")
        total_entity_mapping = []
        for entities in entity_mapping:
            total_entity_mapping.append(entities.split(":"))
        entity_mapping = dict()
        for entity in total_entity_mapping:
            entity_mapping[entity[0]] = entity[1]
        for entities in total_entity_mapping:
            try:
                assert len(entities) == 2
            except AssertionError:
                console.print(
                    f"{RED_START}Entities mapping is given in the wrong format, Right format is{RED_END} "
                    f"{ORANGE3_START}'your_entity_name:ns_entity_name,your_entity_name...'{ORANGE3_END}"
                )
                return False
        return entity_mapping


@click.group(
    name="nlu", short_help="For short text classification and entity extraction"
)
def nlu():
    pass


@nlu.command(name="get-languages")
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_get_languages(log_level: Text):
    setup_logger(log_level=log_level)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(get_languages())
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(
    name="convert-dataset",
    help="To know more about the format and structure kindly refer to"
    "https://docs.neuralspace.ai/natural-language-understanding/concepts/dataset-format-and-converters",
)
@click.option(
    "-F",
    "--from-platform",
    type=click.Choice(["dialogflow", "rasa", "csv", "luis"], case_sensitive=False),
    required=True,
    help="Where to convert data from [dialogflow, rasa, csv, luis]",
)
@click.option(
    "-at",
    "--auto-tag-entities",
    is_flag=True,
    default=False,
    help="If the names of your entities are the same as our pretrained entities "
    "then tag them automatically as pre-trained",
)
@click.option(
    "-L",
    "--language",
    type=click.STRING,
    required=True,
    help="Language in which the dataset is. "
    "Use the following to get a list of supported languages: neuralspace nlu get-languages",
)
@click.option(
    "-em",
    "--entity-mapping",
    type=MappingClass(),
    required=False,
    help="If your entity names don't match with our pre-trained entity names then you can give a mapping like this:"
    "'your_entity_name:ns_entity_name,...' and we will convert them to our format. "
    "This argument works best with the --auto-tag-entities argument as it can happen in a "
    "lot of cases that your entity names don't match our pre-trained entity names, "
    "or even regex and lookup entities. This helps you make make the best out of the platform.",
)
@click.option(
    "-t",
    "--data-type",
    type=click.Choice(["train", "test"], case_sensitive=True),
    default="train",
    help="you can decide whether you want it as test or train data. Defaults to train.",
)
@click.option(
    "-d",
    "--input-path",
    required=True,
    multiple=True,
    help="Name of the folder in which the dataset files are stored. "
    "You can read more about dataset formats here https://docs.neuralspace.ai/natural-language-understanding/concepts/dataset-format-and-converters#dataset-format "  # noqa
    "You can look at sample datasets here: https://github.com/Neural-Space/neuralspace-examples/tree/format_examples/datasets/nlu/Sample%20formats",  # noqa
)
@click.option(
    "-o",
    "--output-path",
    required=True,
    help="Path to the folder in which the converted dataset files will be stored",
)
@click.option(
    "-ime",
    "--ignore-missing-examples",
    is_flag=True,
    default=False,
    help="Will ignore if there are any missing examples in your data (works with CSV format only)",
)
@click.option(
    "-iswp",
    "--ignore-swapped-columns",
    is_flag=True,
    default=False,
    help="Will ignore if you have swapped columns in your nlu examples data (works with csv files only)",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def converter(
    from_platform: Text,
    auto_tag_entities: bool,
    language: Text,
    entity_mapping: typing.Dict[Text, Text],
    input_path: List[Text],
    data_type: Text,
    output_path: Text,
    ignore_missing_examples: bool,
    ignore_swapped_columns: bool,
    log_level: Text,
):
    setup_logger(log_level=log_level)
    if entity_mapping is not False:
        if from_platform == "rasa":
            from neuralspace.nlu.converters.rasaconverter import RasaConverter

            converter = RasaConverter(
                language=language,
                entity_mapping=entity_mapping,
                auto_tag_entities=auto_tag_entities,
            )
            converter.convert(
                input_path=[Path(path) for path in input_path],
                output_directory=Path(output_path),
                dataset_type=DatasetTypes(data_type).value,
            )
        elif from_platform == "dialogflow":
            from neuralspace.nlu.converters.dialogflowconverter import (
                DialogflowConverter,
            )

            converter = DialogflowConverter(
                language=language,
                entity_mapping=entity_mapping,
                auto_tag_entities=auto_tag_entities,
            )
            converter.convert(
                input_path=[Path(path) for path in input_path],
                output_directory=Path(output_path),
                dataset_type=DatasetTypes(data_type).value,
            )
        elif from_platform == "luis":
            from neuralspace.nlu.converters.luisconverter import (
                LuisConverter,
            )

            converter = LuisConverter(
                language=language,
                entity_mapping=entity_mapping,
                auto_tag_entities=auto_tag_entities,
            )
            converter.convert(
                input_path=[Path(path) for path in input_path],
                output_directory=Path(output_path),
                dataset_type=DatasetTypes(data_type).value,
            )
        else:
            from neuralspace.nlu.converters.csvconverter import CsvConverter

            converter = CsvConverter(
                language=language,
                entity_mapping=entity_mapping,
                auto_tag_entities=auto_tag_entities,
                ignore_missing_examples=ignore_missing_examples,
                ignore_swapped_columns=ignore_swapped_columns,
            )
            converter.convert(
                input_path=Path(input_path[0]),
                output_directory=Path(output_path),
                dataset_type=DatasetTypes(data_type).value,
            )
        print_converted_data_inference(from_platform, Path(output_path), converter)
    else:
        console.print(
            f"{ERROR} There is some error in the entity mapping you have given"
        )


@nlu.command(
    name="convert-huggingface-ner-dataset",
)
@click.option(
    "-hf",
    "--huggingface-dataset",
    required=True,
    type=click.STRING,
    help="Exact name of the NER dataset from huggingface dataset hub",
)
@click.option(
    "-s",
    "--subset",
    required=True,
    type=click.STRING,
    multiple=True,
    help="Go to the dataset page on huggingface and pass all the subsets you want as shown in the dataset preview."
    "For multiple subsets, pass them in one string seperated by space",
)
@click.option(
    "-L",
    "--language",
    type=click.STRING,
    required=True,
    help="Language in which the dataset is. "
    "Use the following to get a list of supported languages: neuralspace nlu get-languages"
    "For datasets having more than one language, select 'multilingual'",
)
@click.option(
    "-tr",
    "--num-train-examples",
    type=click.INT,
    required=False,
    default=-1,
    help="Pass the number of examples you want in your converted training dataset for each subset"
    "Dont set this parameter if you want to select all the examples available in the dataset",
)
@click.option(
    "-te",
    "--num-test-examples",
    type=click.INT,
    required=False,
    default=-1,
    help="Pass the number of examples you want in your converted test dataset for each subset"
    "Dont set this parameter if you want to select all the examples available in the dataset",
)
@click.option(
    "-o",
    "--output-path",
    required=True,
    type=click.STRING,
    help="Path to the folder in which the converted dataset files will be stored",
)
@click.option(
    "-me",
    "--max-entities",
    required=False,
    default=30,
    type=click.INT,
    help="Only keeps sentences having number of entities less than selected value."
    "Sentences having more entities are discarded.",
)
def huggingface_converter(
    huggingface_dataset: Text,
    subset: List,
    language: Text,
    num_train_examples: int,
    num_test_examples: int,
    output_path: Text,
    max_entities: int,
):
    from neuralspace.nlu.converters.huggingface_nerconverter import (
        HuggingfaceNerConverter,
    )

    converter = HuggingfaceNerConverter(
        language=language, max_entities_in_text=max_entities
    )
    converter.convert(
        hf_dataset_name=huggingface_dataset,
        subsets=list(subset),
        num_train_examples=num_train_examples,
        num_test_examples=num_test_examples,
        output_directory=Path(output_path),
    )


@nlu.command(name="create-project")
@click.option(
    "-p",
    "--project-name",
    type=click.STRING,
    required=False,
    help="Project Name",
)
@click.option(
    "-L",
    "--languages",
    type=click.STRING,
    multiple=True,
    required=True,
    help="Language code that you want to select for this project. To get the list of supported languages try:"
    " neuralspace nlu get-languages",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_create_project(project_name: Text, languages: List[Text], log_level: Text):
    setup_logger(log_level=log_level)
    languages = list(languages)
    if project_name is None:
        project_name = randomname.get_name(**NAME_GENERATOR_CONFIG)
    if not languages:
        ValueError(f"Languages is mandatory. You can select from {SUPPORTED_LANGUAGES}")
    loop = asyncio.new_event_loop()
    loop.run_until_complete(create_project(project_name, languages))
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(name="delete-project")
@click.option(
    "-p",
    "--project-id",
    type=click.STRING,
    required=True,
    help="id of the project you want to delete",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_delete_project(project_id: Text, log_level: Text):
    setup_logger(log_level=log_level)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(delete_project(project_id))
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(name="list-projects")
@click.option(
    "-S",
    "--search",
    type=click.STRING,
    required=False,
    default="",
    help="Substring search keyword",
)
@click.option(
    "-p",
    "--page-number",
    type=click.INT,
    required=False,
    default=1,
    help="Which page number to fetch",
)
@click.option(
    "-s",
    "--page-size",
    type=click.INT,
    required=False,
    default=20,
    help="items per page to fetch",
)
@click.option(
    "-L",
    "--language",
    type=click.STRING,
    multiple=True,
    default=[],
    help="List of languages. E.g., -L en -L de; You can add multiple languages",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Verbose results",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_list_projects(
    search: Text,
    page_number: int,
    page_size: int,
    language: List[Text],
    verbose: bool,
    log_level: Text,
):
    setup_logger(log_level=log_level)
    languages = list(language)
    if not languages:
        ValueError(f"Languages is mandatory. You can select from {SUPPORTED_LANGUAGES}")
    loop = asyncio.new_event_loop()
    loop.run_until_complete(
        list_projects(search, page_size, page_number, languages, verbose)
    )
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(name="upload-dataset")
@click.option(
    "-d",
    "--dataset-file",
    type=click.Path(exists=True, dir_okay=False),
    required=True,
    help="An nlu dataset file in NeuralSpace format. "
    "To convert from other platforms like Rasa, Dialogflow, etc. try: ",
)
@click.option(
    "-p",
    "--project-id",
    type=click.STRING,
    required=True,
    help="NLU project id",
)
@click.option(
    "-L",
    "--language",
    type=click.Choice(SUPPORTED_LANGUAGES),
    required=True,
    help="Language",
)
@click.option(
    "-s",
    "--skip-first",
    type=click.INT,
    required=False,
    default=0,
    help="Skip this many example from the beginning. "
    "Use this in case you have to interrupt upload or if some error occurred while uploading.",
)
@click.option(
    "-e",
    "--ignore-errors",
    type=click.BOOL,
    required=False,
    default=False,
    help="To ignore errors and go to the next example",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_upload_dataset(
    dataset_file: Path,
    project_id: Text,
    language: Text,
    skip_first: int,
    ignore_errors: bool,
    log_level: Text,
):
    setup_logger(log_level=log_level)
    dataset = json.loads(Path(dataset_file).read_text())
    loop = asyncio.new_event_loop()
    loop.run_until_complete(
        upload_dataset(
            dataset,
            project_id=project_id,
            language=language,
            skip_first=skip_first,
            ignore_errors=ignore_errors,
        )
    )
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(name="list-examples")
@click.option(
    "-p",
    "--project-id",
    type=click.STRING,
    required=True,
    help="Project ID",
)
@click.option(
    "-L",
    "--language",
    type=click.STRING,
    required=True,
    help="Language",
)
@click.option(
    "-P",
    "--prepared",
    type=click.Choice(["true", "false"]),
    required=False,
    default="true",
    help="Flag to filter only prepared examples",
)
@click.option(
    "-t",
    "--type",
    type=click.Choice(["train", "test"]),
    required=False,
    default="train",
    help="Flag to filter only train or test examples",
)
@click.option(
    "-i",
    "--intent",
    type=click.STRING,
    required=False,
    help="Flag to filter examples only for this intent",
)
@click.option(
    "-n",
    "--page-number",
    type=click.INT,
    required=False,
    default=1,
    help="Which page number to fetch",
)
@click.option(
    "-s",
    "--page-size",
    type=click.INT,
    required=False,
    default=20,
    help="items per page to fetch",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Verbose results",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_list_examples(
    project_id: Text,
    language: Text,
    prepared: Text,
    type: Text,
    intent: Text,
    page_number: int,
    page_size: int,
    verbose: bool,
    log_level: Text,
):
    setup_logger(log_level=log_level)
    prepared_flag = False
    if prepared == "true":
        prepared_flag = True
    loop = asyncio.new_event_loop()
    loop.run_until_complete(
        list_examples(
            project_id,
            language,
            prepared_flag,
            type,
            intent,
            page_number,
            page_size,
            verbose,
        )
    )
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(name="delete-example")
@click.option(
    "-e",
    "--example-id",
    type=click.STRING,
    required=True,
    multiple=True,
    help="Example ID",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_delete_example(example_id: List[Text], log_level: Text):
    setup_logger(log_level=log_level)
    example_ids = list(example_id)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(delete_examples(example_ids))
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(name="train")
@click.option(
    "-p",
    "--project-id",
    type=click.STRING,
    required=True,
    help="Project ID",
)
@click.option(
    "-L",
    "--language",
    type=click.STRING,
    required=True,
    help="Language",
)
@click.option(
    "-m",
    "--model-name",
    type=click.STRING,
    required=False,
    help="Name of the model",
)
@click.option(
    "-n",
    "--training-jobs",
    type=click.INT,
    required=False,
    default=1,
    help="No of training jobs to be queued",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_train_model(
    project_id: Text,
    language: Text,
    model_name: Text,
    log_level: Text,
    training_jobs: int,
):
    setup_logger(log_level=log_level)
    print_logo()
    if model_name is None:
        model_name = randomname.get_name(**NAME_GENERATOR_CONFIG)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(
        train_model(project_id, language, model_name, training_jobs)
    )
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(name="model-status")
@click.option(
    "-m",
    "--model-id",
    type=click.STRING,
    required=True,
    help="Model ID",
)
@click.option(
    "-w",
    "--wait",
    is_flag=True,
    default=True,
    help="Wait for training to complete",
)
@click.option(
    "-i",
    "--wait-interval",
    type=click.INT,
    required=False,
    default=1,
    help="Time to wait between consecutive status-check pools.",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_check_model_status(
    model_id: Text, wait: bool, wait_interval: int, log_level: Text
):
    setup_logger(log_level=log_level)
    print_logo()
    loop = asyncio.new_event_loop()
    loop.run_until_complete(wait_till_training_completes(model_id, wait, wait_interval))
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(name="list-models")
@click.option(
    "-p",
    "--project-id",
    type=click.STRING,
    required=True,
    help="Project ID",
)
@click.option(
    "-L",
    "--language",
    type=click.STRING,
    required=True,
    help="Language",
)
@click.option(
    "-s",
    "--training-status",
    type=click.STRING,
    required=False,
    help=f"Flag to filter models with the given training status. "
    f"These are some valid values: {', '.join(TRAINING_PROGRESS)}",
)
@click.option(
    "-n",
    "--page-number",
    type=click.INT,
    required=False,
    default=1,
    help="Which page number to fetch",
)
@click.option(
    "-s",
    "--page-size",
    type=click.INT,
    required=False,
    default=20,
    help="items per page to fetch",
)
@click.option(
    "-v",
    "--verbose",
    is_flag=True,
    help="Verbose results",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_list_models(
    project_id: Text,
    language: Text,
    training_status: List[Text],
    page_number: int,
    page_size: int,
    verbose: bool,
    log_level: Text,
):
    setup_logger(log_level=log_level)
    if training_status:
        training_status = list(training_status)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(
        list_models(
            project_id, language, training_status, page_number, page_size, verbose
        )
    )
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(name="delete-model")
@click.option(
    "-m",
    "--model-id",
    type=click.STRING,
    required=True,
    help="Model ID",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_delete_model(model_id: Text, log_level: Text):
    setup_logger(log_level=log_level)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(delete_models(model_id))
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(name="deploy")
@click.option(
    "-m",
    "--model-id",
    type=click.STRING,
    required=True,
    help="Model ID",
)
@click.option(
    "-n",
    "--n-replicas",
    type=click.INT,
    required=False,
    default=1,
    help="Number of replicas to deploy for this model. One Replica gives you 5-8 requests per second throughput.",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_deploy(model_id: Text, n_replicas: int, log_level: Text):
    setup_logger(log_level=log_level)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(deploy(model_id, n_replicas))
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(name="parse")
@click.option(
    "-m",
    "--model-id",
    type=click.STRING,
    required=True,
    help="Model ID",
)
@click.option(
    "-i",
    "--input-text",
    type=click.STRING,
    required=True,
    help="Text to parse",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def nlu_parse(model_id: Text, input_text: Text, log_level: Text):
    setup_logger(log_level=log_level)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(parse(model_id, input_text))
    loop.run_until_complete(get_async_http_session().close())


@nlu.command(name="get-pretrained-entities")
@click.option("-L", "--language", type=click.STRING, required=True)
@click.option("-s", "--search", type=click.STRING, default="")
@click.option("-p", "--page-number", type=click.INT, default=1)
@click.option("-P", "--page-size", type=click.INT, default=5)
@click.option(
    "-T", "--entity-type", type=click.Choice(NER_TYPES), default="pre-trained"
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def get_pretrained_entities(
    language: Text,
    search: Text,
    page_number: int,
    page_size: int,
    entity_type: Text,
    log_level: Text,
):
    setup_logger(log_level=log_level)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(
        get_entity_list(language, search, page_number, page_size, entity_type)
    )
    loop.run_until_complete(get_async_http_session().close())



import json
from pathlib import Path
from typing import Any, Dict, Text

from rich.table import Table

from neuralspace.constants import (
    KEY_MODEL_IDS,
    KEY_PATH_TO_PROJECT_ID,
    KEY_PROJECT_DETAILS,
    KEY_PROJECT_IDS,
    KEY_RASA,
    NS_RASA_METADATA_FILE,
    ORANGE3_END,
    ORANGE3_START,
    neuralspace_home,
)
from neuralspace.nlu.apis import create_project_id


def metadata_file_path() -> Path:
    metadata_path = neuralspace_home() / KEY_RASA / NS_RASA_METADATA_FILE
    metadata_path.parent.mkdir(exist_ok=True)
    if not metadata_path.exists():
        initial_meta_file = {
            KEY_PROJECT_IDS: [],
            KEY_PROJECT_DETAILS: {},
            KEY_PATH_TO_PROJECT_ID: {},
        }
        with open(metadata_path, "w") as meta_file_pointer:
            json.dump(initial_meta_file, meta_file_pointer, indent=4)
    return metadata_path


def get_metadata() -> Dict[Text, Any]:
    with open(metadata_file_path(), "r") as file_pointer:
        metadata = json.load(file_pointer)
    return metadata


def update_metadata(metadata: Dict[Text, Any]):
    with open(metadata_file_path(), "w") as file_pointer:
        json.dump(metadata, file_pointer, ensure_ascii=False, indent=4)


def get_project_id_if_exists(
    working_directory: Text, create_if_doesnt_exist: bool = False, language: Text = None
) -> Text:
    metadata = get_metadata()
    project_id = None
    if working_directory in metadata[KEY_PATH_TO_PROJECT_ID]:
        project_id = metadata[KEY_PATH_TO_PROJECT_ID][working_directory]
    elif create_if_doesnt_exist:
        project_id = create_project_id(language=language)
    return project_id


def get_recent_model_id(project_id: Text):
    meta_data = get_metadata()
    return meta_data[KEY_PROJECT_DETAILS][project_id][KEY_MODEL_IDS][-1]


def print_table_description_for_examples(
    total_uploaded_examples,
    total_removed_examples,
    total_uploaded_entities,
    total_removed_entities,
    total_removed_entities_examples,
    total_uploaded_entities_examples,
) -> Table:
    table = Table(show_header=True, header_style="bold #c47900", show_lines=True)
    table.add_column("Contents")
    table.add_column("Total")
    table.add_row(
        f"{ORANGE3_START}Uploaded NLU Examples{ORANGE3_END}",
        str(total_uploaded_examples),
    )
    table.add_row(
        f"{ORANGE3_START}Uploaded Entities{ORANGE3_END}", str(total_uploaded_entities)
    )
    table.add_row(
        f"{ORANGE3_START}Uploaded Entities Example{ORANGE3_END}",
        str(total_uploaded_entities_examples),
    )
    table.add_row(
        f"{ORANGE3_START}Removed NLU Examples{ORANGE3_END}", str(total_removed_examples)
    )
    table.add_row(
        f"{ORANGE3_START}Removed Entities{ORANGE3_END}", str(total_removed_entities)
    )
    table.add_row(
        f"{ORANGE3_START}Removed Entities Example{ORANGE3_END}",
        str(total_removed_entities_examples),
    )
    return table



from pathlib import Path
from typing import Any, Dict, List, Optional, Text, Tuple

import pandas as pd

from neuralspace.constants import (
    ENTITIES,
    ENTITY,
    ENTITY_TYPE,
    EXAMPLES,
    INTENT,
    LANGUAGE,
    LOOKUP_FILE,
    NLU_FILE,
    REGEX,
    REGEX_FILE,
    SYNONYM,
    SYNONYMS_FILE,
    TEXT,
)
from neuralspace.datamodels import DatasetType
from neuralspace.nlu.converters.rasaconverter import RasaConverter


class NLUValidationError(Exception):
    pass


class EntityValidationError(Exception):
    pass


class CsvConverter(RasaConverter):
    def __init__(
        self,
        entity_mapping: Optional[Dict[Text, Text]],
        auto_tag_entities: bool,
        language: Text,
        ignore_missing_examples: bool,
        ignore_swapped_columns: bool,
    ):
        self.ignore_missing_examples = ignore_missing_examples
        self.ignore_swapped_columns = ignore_swapped_columns
        super(CsvConverter, self).__init__(
            language=language,
            entity_mapping=entity_mapping,
            auto_tag_entities=auto_tag_entities,
        )

    def __validate_nlu_file(self, dataset: pd.DataFrame):
        if not len(dataset.columns) == 2:
            raise NLUValidationError(
                "Two Columns(label, text) are required to convert the dataset"
            )
        else:
            if not self.ignore_swapped_columns:
                if len(dataset[dataset.columns[0]].unique()) >= len(dataset):
                    raise NLUValidationError(
                        "Check if you have given the labels in first column and texts in second column"
                        " if you have given it correct then run the command with -iswp"
                    )
        if dataset[dataset.columns[1]].isna().any().any():
            raise NLUValidationError("Found missing Text in your csv data")
        if not self.ignore_missing_examples and dataset.isna().any().any():
            raise NLUValidationError("Found missing label data in your csv")

    @staticmethod
    def __validate_entity_file(entities: pd.DataFrame):
        for entity_label, examples in entities.items():
            if len(examples.dropna().to_list()) < 1:
                raise EntityValidationError(
                    f"There is no examples present for entity: {entity_label}"
                )

    def nlu_data_converter(self, dataset: pd.DataFrame) -> List[Dict[Text, Any]]:

        import rasa.shared.nlu.training_data.entities_parser as entities_parser

        """

        +--------+------+
        | Label  | Text |
        +--------+------+
        |A_label | sdf  |
        +---------------+

        :return: training data in the Neuralspace format
        """
        nlu_data = []
        self.__validate_nlu_file(dataset)
        labels = dataset[dataset.columns[0]]
        texts = dataset[dataset.columns[1]]
        for label, text in zip(labels, texts):
            nlu_data.append(
                {
                    TEXT: entities_parser.replace_entities(text),
                    INTENT: label,
                    ENTITIES: entities_parser.find_entities_in_training_example(text),
                }
            )
        return nlu_data

    def synonym_converter(self, synonyms: pd.DataFrame) -> List[Dict[Text, Any]]:
        self.__validate_entity_file(synonyms)
        synonym_data = []
        for synonym_label, examples in synonyms.items():
            synonym_data.append(
                {
                    ENTITY: synonym_label,
                    EXAMPLES: examples.dropna().to_list(),
                    ENTITY_TYPE: SYNONYM,
                    LANGUAGE: self.language,
                }
            )
        return synonym_data

    def regex_converter(self, entities: pd.DataFrame) -> List[Dict[Text, Any]]:
        self.__validate_entity_file(entities)
        regex_data = []
        for regex_label, examples in entities.items():
            regex_data.append(
                {
                    ENTITY: regex_label,
                    EXAMPLES: examples.dropna().to_list(),
                    ENTITY_TYPE: REGEX,
                    LANGUAGE: self.language,
                }
            )
        return regex_data

    def lookup_converter(self, entities: pd.DataFrame) -> List[Dict[Text, Any]]:
        self.__validate_entity_file(entities)
        lookup_data = []
        for lookup_label, examples in entities.items():
            lookup_data.append(
                {
                    ENTITY: lookup_label,
                    EXAMPLES: examples.dropna().to_list(),
                    ENTITY_TYPE: SYNONYM,
                    LANGUAGE: self.language,
                }
            )
        return lookup_data

    @staticmethod
    def get_data_files(
        directory_path: Path,
    ) -> Tuple[pd.DataFrame, pd.DataFrame, pd.DataFrame, pd.DataFrame]:
        return (
            pd.read_csv(directory_path / NLU_FILE)
            if (directory_path / NLU_FILE).exists()
            else pd.DataFrame(),
            pd.read_csv(directory_path / SYNONYMS_FILE)
            if (directory_path / SYNONYMS_FILE).exists()
            else pd.DataFrame(),
            pd.read_csv(directory_path / REGEX_FILE)
            if (directory_path / REGEX_FILE).exists()
            else pd.DataFrame(),
            pd.read_csv(directory_path / LOOKUP_FILE)
            if (directory_path / LOOKUP_FILE).exists()
            else pd.DataFrame(),
        )

    def convert(
        self, input_path: Path, output_directory: Path, dataset_type: DatasetType
    ):
        nlu_data, synonym_data, regex_data, lookup_data = self.get_data_files(
            input_path
        )
        if nlu_data.columns.to_list():
            self.nlu_data = self.nlu_data_converter(nlu_data)
        if synonym_data.columns.to_list():
            self.synonym_data = self.synonym_converter(synonym_data)
        if lookup_data.columns.to_list():
            self.lookup_data = self.lookup_converter(lookup_data)
        if regex_data.columns.to_list():
            self.regex_data = self.regex_converter(regex_data)
        self.post_process(dataset_type=dataset_type, output_directory=output_directory)



import asyncio
from typing import Text

import click
from rich.console import Console
from rich.prompt import Prompt
from rich.table import Table

from neuralspace import VERSION
from neuralspace.apis import get_async_http_session
from neuralspace.augmentation.commands import augmentation
from neuralspace.language_detection.commands import language_detection
from neuralspace.ner.commands import ner
from neuralspace.nlu.commands import nlu
from neuralspace.translation.commands import translation
from neuralspace.transliteration.commands import transliteration
from neuralspace.utils import (
    add_apps_to_table,
    add_heading_to_description_table,
    app_install,
    do_login,
    print_logo,
    print_logo_and_description,
    setup_logger,
)

console = Console()


def close_connection():
    loop = asyncio.new_event_loop()
    loop.run_until_complete(get_async_http_session().close())


# atexit.register(close_connection)


class RichGroup(click.Group):
    def format_help(self, ctx, formatter):
        print_logo_and_description()


@click.group(cls=RichGroup)
@click.version_option(VERSION)
def entrypoint():
    pass


@entrypoint.command(name="list-apps")
def list_apps():
    table = Table(show_header=True, header_style="bold #c47900", show_lines=True)
    table = add_heading_to_description_table(table)
    table = add_apps_to_table(table)
    console.print(table)


@entrypoint.command(name="install-app")
@click.option(
    "-n",
    "--name",
    type=click.STRING,
    required=True,
    help="Code of the app you want to install. To get App codes try: neuralspace list-apps",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def install_app(name: Text, log_level: Text):
    setup_logger(log_level=log_level)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(app_install(name))
    loop.run_until_complete(get_async_http_session().close())


@entrypoint.command(name="login", short_help="To login to the neuralSpace platform")
@click.option(
    "-e",
    "--email",
    type=click.STRING,
    required=False,
    help="Your NeuralSpace email id",
)
@click.option(
    "-p",
    "--password",
    type=click.STRING,
    required=False,
    help="Your NeuralSpace account password",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def login(email: Text, password: Text, log_level: Text):
    if not email:
        email = Prompt.ask("Enter your email")
    if not password:
        password = Prompt.ask("Enter your password", password=True)
    setup_logger(log_level=log_level)
    print_logo()
    loop = asyncio.new_event_loop()
    loop.run_until_complete(do_login(email, password))
    loop.run_until_complete(get_async_http_session().close())


entrypoint.add_command(nlu)
entrypoint.add_command(ner)
entrypoint.add_command(translation)
entrypoint.add_command(transliteration)
entrypoint.add_command(augmentation)
entrypoint.add_command(language_detection)

if __name__ == "__main__":
    entrypoint()



import asyncio
from typing import Text

import click

from neuralspace.apis import get_async_http_session
from neuralspace.language_detection.apis import (
    get_language_detection_response,
    get_languages,
)
from neuralspace.utils import setup_logger


@click.group(name="language-detection")
def language_detection():
    pass


@language_detection.command(
    name="parse",
)
@click.option(
    "-t",
    "--text",
    type=click.STRING,
    required=True,
    help="text to be processed",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def detect_language(text: Text, log_level: Text):
    setup_logger(log_level=log_level)
    if not text:
        ValueError("Please enter the text for which you wish to detect language")
    loop = asyncio.new_event_loop()
    loop.run_until_complete(get_language_detection_response(text))
    loop.run_until_complete(get_async_http_session().close())


@language_detection.command(name="get-languages")
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def language_detection_get_languages(log_level: Text):
    setup_logger(log_level=log_level)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(get_languages())
    loop.run_until_complete(get_async_http_session().close())



import asyncio
from typing import Text

import click

from neuralspace.apis import get_async_http_session
from neuralspace.constants import NER_TYPES
from neuralspace.ner.apis import get_ner_response
from neuralspace.ner.constants import SUPPORTED_LANGUAGES
from neuralspace.utils import get_entity_list, setup_logger


@click.group(name="ner")
def ner():
    pass


@ner.command(
    name="parse",
)
@click.option(
    "-t",
    "--text",
    type=click.STRING,
    required=True,
    help="text to be parsed",
)
@click.option(
    "-L",
    "--language",
    type=click.STRING,
    required=True,
    help="language that the text was in",
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def parse(text: Text, language: Text, log_level: Text):
    setup_logger(log_level=log_level)
    if not language:
        ValueError(f"Language is mandatory. You can select from {SUPPORTED_LANGUAGES}")
    if not text:
        ValueError("Please enter the text to parse")
    loop = asyncio.new_event_loop()
    loop.run_until_complete(get_ner_response(text, language))
    loop.run_until_complete(get_async_http_session().close())


@ner.command(name="get-pretrained-entities")
@click.option("-L", "--language", type=click.STRING, required=True)
@click.option("-s", "--search", type=click.STRING, default="")
@click.option("-p", "--page-number", type=click.INT, default=1)
@click.option("-P", "--page-size", type=click.INT, default=5)
@click.option(
    "-T", "--entity-type", type=click.Choice(NER_TYPES), default="pre-trained"
)
@click.option(
    "-l", "--log-level", type=click.Choice(["INFO", "DEBUG", "ERROR"]), default="INFO"
)
def get_pretrained_entities(
    language: Text,
    search: Text,
    page_number: int,
    page_size: int,
    entity_type: Text,
    log_level: Text,
):
    setup_logger(log_level=log_level)
    loop = asyncio.new_event_loop()
    loop.run_until_complete(
        get_entity_list(language, search, page_number, page_size, entity_type)
    )
    loop.run_until_complete(get_async_http_session().close())



import json
import logging
import os
from copy import copy
from pathlib import Path
from typing import Any, Dict, List, Text

from rich.console import Console
from rich.logging import RichHandler
from rich.table import Table

from neuralspace.apis import get_async_http_session
from neuralspace.constants import (
    APP_IS_INSTALLED,
    AUTHORIZATION,
    BOLD_END,
    BOLD_START,
    BOOK,
    COMMON_HEADERS,
    CONFIDENCE,
    COUNT,
    CROSS,
    DARK_ORANGE_END,
    DARK_ORANGE_START,
    DATA,
    DESCRIPTION,
    DOOR,
    END_INDEX,
    ENTITIES,
    ENTITY,
    ENTITY_TYPE,
    FAST_FORWARD,
    FILTER,
    FROM,
    GREEN_TICK,
    INFO,
    INSTALL_APP_COMMAND,
    INSTALL_APP_URL,
    KEY,
    LANGUAGE,
    LIST_ENTITY_URL,
    LOGIN_URL,
    LOGO_HASH,
    ORANGE3_END,
    ORANGE3_START,
    PAGE_NUMBER,
    PAGE_SIZE,
    SEARCH,
    START_INDEX,
    TIME,
    TO,
    TYPE,
    VALUE,
    auth_path,
    neuralspace_url,
)

logger = logging.getLogger("rich")
console = Console()


def setup_logger(log_level: Text):
    global logger
    logging.basicConfig(level=log_level, handlers=[RichHandler(level=logging.INFO)])
    if log_level == "INFO":
        logger.level = logging.INFO
        logging.getLogger("asyncio").setLevel(logging.INFO)
    elif log_level == "DEBUG":
        logger.level = logging.DEBUG
        logging.getLogger("asyncio").setLevel(logging.DEBUG)
    elif log_level == "ERROR":
        logger.level = logging.ERROR
        logging.getLogger("asyncio").setLevel(logging.ERROR)


def get_logo():
    logo_path = (Path(os.path.realpath(__file__))).parent / "data" / "logo.txt"
    return f"{logo_path.read_text()}"


def print_logo():
    logo_path = (Path(os.path.realpath(__file__))).parent / "data" / "logo.txt"
    console.print(
        f"\n{BOLD_START}{logo_path.read_text()}{BOLD_END}", style=f"{LOGO_HASH}"
    )


def register_auth_token(login_response: Dict[Text, Text]):
    if "data" in login_response and "auth" in login_response["data"]:
        with open(str(auth_path()), "w") as f:
            json.dump(login_response, f)
    else:
        raise ValueError(f"Login response is malformed: {login_response}")


def get_auth_token() -> Text:
    if auth_path().exists():
        credentials = json.loads(auth_path().read_text())
        return credentials["data"]["auth"]
    else:
        raise FileNotFoundError(
            "Credentials file not found. Consider logging in using."
            " Seems like you have not logged in. "
            "`neuralspace login --email <your-neuralspace-email-id> "
            "--password <your-password>`"
        )


def is_success_status(status_code: int) -> bool:
    success = False
    if 200 <= status_code < 300:
        success = True
    return success


async def do_login(email: Text, password: Text):
    user_data = {"email": email, "password": password}
    logger.debug(f"Login{DOOR} attempt for: {user_data}")
    async with get_async_http_session().post(
        url=f"{neuralspace_url()}/{LOGIN_URL}",
        data=json.dumps(user_data),
        headers=COMMON_HEADERS,
    ) as response:
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            register_auth_token(json_response)
            console.print(f"> {INFO} {GREEN_TICK}  Login successful!")
            console.print(f"> {INFO} {KEY} Credentials registered at {auth_path()}")
            console.print(
                f"{FAST_FORWARD}  Install an app: [dark_orange3]{INSTALL_APP_COMMAND}[/dark_orange3]"
            )
        else:
            console.print(
                f"> [red]ERROR[/red] {CROSS} Login failed! Please check your username and password"
            )


async def app_install(name: Text):
    user_data = {"appType": name}
    logger.debug(f"Installing app: {name}")
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    async with get_async_http_session().post(
        url=f"{neuralspace_url()}/{INSTALL_APP_URL}",
        data=json.dumps(user_data),
        headers=HEADERS,
    ) as response:
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            console.print(f"> {INFO} {GREEN_TICK} Install Successful!")
            console.print(
                f"{FAST_FORWARD} To use the {name} app: {DARK_ORANGE_START}neuralspace {name}{DARK_ORANGE_END} --help"
            )
        else:
            if json_response["message"] == APP_IS_INSTALLED:
                console.print(f"> {INFO} {GREEN_TICK}  {APP_IS_INSTALLED}")
                console.print(
                    f"{FAST_FORWARD} To use the {name} app: {DARK_ORANGE_START}"
                    f"neuralspace {name}{DARK_ORANGE_END} --help"
                )

            else:
                console.print(f"> [red]ERROR[/red] {CROSS} Install Failed!")
                console.print(
                    f"\n{json.dumps(json_response, indent=4, ensure_ascii=False)}"
                )


def get_list_chunks(items: List[Any], chunk_size: int):
    for chunk in range(0, len(items), chunk_size):
        yield items[chunk : chunk + chunk_size]  # noqa : E203


def add_apps_to_table(table: Table) -> Table:
    table.add_row(
        f"{BOLD_START}NeuraLingo{BOLD_END}",
        f"{BOLD_START}nlu{BOLD_END}",
        "neuralspace [bold red]nlu[/bold red]",
        "Whether you are using chatbots, voicebots, or process automation engines, they are all powered by Natural "
        "Language Understanding (NLU). It's main purpose is to "
        "understand the intent of the user, and extract relevant "
        "information (entities) from what they said (speech) or wrote (text) to perform a relevant action. "
        "Entities can be anything from names, addresses, account numbers to very domain specific terms like names of "
        "chemicals, medicines, etc. Sometimes it also predicts the sentiment of the user which helps the bot respond "
        "to the user in a more empathetic tone.",
    )
    table.add_row(
        f"{BOLD_START}Entity Recognition{BOLD_END}",
        f"{BOLD_START}ner{BOLD_END}",
        "neuralspace [bold red]ner[/bold red]",
        "Entities play a major role in language understanding. To perform an action on a certain user query you not "
        "only need to understand the intent behind it but also the entities present in it. "
        "E.g., if someone says 'flights from Berlin to London', the intent here is flight-search and entities are "
        "Berlin and London, which are of type city. In a given piece of text, entities can be anything from names, "
        "addresses, account numbers to very domain specific terms like names of chemicals, medicines, etc. "
        "Essentially any valuable information that can be extracted from text.",
    )
    table.add_row(
        f"{BOLD_START}Machine Translation{BOLD_END}",
        f"{BOLD_START}translation{BOLD_END}",
        "neuralspace [bold red]translation[/bold red]",
        "Whether we are talking about subtitles, government documents or question papers for exams, all of them "
        "need to be translated into multiple languages. Manually translating documents at such a scale is not "
        "only expensive but also an extremely time consuming process. With the help of Neural Machine "
        "Translation (NMT) you can drastically reduce the amount of time it takes for manual translation "
        "of documents. Our translation models are all state-of-the-art Artificial Neural Networks which can "
        "translate text in over 100 languages. ",
    )
    table.add_row(
        f"{BOLD_START}Transliteration{BOLD_END}",
        f"{BOLD_START}transliteration{BOLD_END}",
        "neuralspace [bold red]transliteration[/bold red]",
        "For languages that don't use the latin script, e.g., Arabic, Hindi, Punjabi, Sinhala and many "
        "other spoken around the world, typing can be challenging as keyboards/keypads mostly default to "
        "latin characters. That makes creating content in vernacular languages difficult. With "
        "transliteration you can create content in these languages using your latin keypad. "
        "For instance, you type a word on the latin keypad the way you would pronounce it in Punjabi, "
        "and using transliteration you can convert that into Punjabi. It transforms a word from one "
        "alphabet to the other phonetically. ",
    )
    table.add_row(
        f"{BOLD_START}NeuralAug{BOLD_END}",
        f"{BOLD_START}augmentation{BOLD_END}",
        "neuralspace [bold red]augmentation[/bold red]",
        "Any language processing task requires data, and we all wish we could generate data "
        "magically. That was exactly the idea behind building NeuralAug. Given a sentence NeuralAug "
        "can generate up to ten sentences keeping the intent of the original sentence intact. It "
        "can help in creating datasets faster and make language processing models more robust. ",
    )
    table.add_row(
        f"{BOLD_START}Language Detection{BOLD_END}",
        f"{BOLD_START}language-detection{BOLD_END}",
        "neuralspace [bold red]language-detection[/bold red]",
        "Dealing with different languages poses a fundamental problem of first detecting what language "
        "you are looking at. Language-detection service will help you to detect the language based on "
        "your input text with given confidence score. Use it with other APIs like transliteration "
        "or translation and make your NLP application cater to any language you want.",
    )
    return table


def add_heading_to_description_table(table: Table) -> Table:
    table.add_column("App Name", style="#c47900")
    table.add_column("App Code", style="#c47900")
    table.add_column("Command", style="#c47900")
    table.add_column("Description")
    return table


def print_logo_and_description():
    console.print(f"[bold]{get_logo()}[/bold]", style="#c47900")
    console.print(
        "[bold magenta]Website: [/bold magenta][link]https://neuralspace.ai[/link]"
    )
    console.print(
        "[bold magenta]Docs: [/bold magenta][link]https://docs.neuralspace.ai[/link]"
    )
    console.print(
        "[bold magenta]Platform Login: [/bold magenta][link]https://platform.neuralspace.ai[/link]"
    )
    console.print("[bold magenta]Commands: [/bold magenta]")
    table = Table(show_header=True, header_style="bold #c47900", show_lines=True)
    table = add_heading_to_description_table(table)
    table.add_row(
        "[bold]Login[/bold]",
        "neuralspace [bold red]login[/bold red]",
        "Login to the our platform and save credentials locally",
    )
    table.add_row(
        "[bold]List Apps[/bold]",
        "neuralspace [bold red]list-apps[/bold red]",
        "List all the available apps in our platform and their respective codes",
    )
    table.add_row(
        "[bold]Install App[/bold]",
        "neuralspace [bold red]install-app[/bold red]",
        "Installs the specific app on your account",
    )
    table = add_apps_to_table(table)
    console.print(table)


def print_ner_response(list_of_entities: List[Dict[Text, Text]], original_text: Text):
    if not list_of_entities:
        console.print(
            f"There is no entities in your text: [bold red]{original_text}[/bold red]"
        )
    else:
        table = Table(show_header=True, header_style="orange3")
        table.add_column("idx", style="sandy_brown")
        table.add_column("Type", style="green")
        table.add_column("Value", style="green")
        table.add_column("Entities marked in original text")
        for i, entities in enumerate(list_of_entities):
            start_index = entities[START_INDEX]
            end_index = entities[END_INDEX]
            marked_sentence = ""
            for idx, character in enumerate(original_text):
                if idx == start_index:
                    marked_sentence += "[bold green]"
                elif idx == end_index:
                    marked_sentence += "[/bold green]"
                marked_sentence += character
            if entities[TYPE] == TIME and isinstance(entities[VALUE], dict):
                type_time_format = (
                    f"From: {str(entities[VALUE][FROM])} To: {str(entities[VALUE][TO])}"
                )
                table.add_row(str(i), entities[TYPE], type_time_format, marked_sentence)
            else:
                table.add_row(str(i), entities[TYPE], entities[VALUE], marked_sentence)
        console.print(table)


def print_translation_response(
    text: Text, src_language: Text, tgt_language: Text, translated_text: Text
):
    table = Table(show_header=True, header_style="orange3")
    table.add_column("Text", style="sandy_brown")
    table.add_column("Src Language", style="green")
    table.add_column("Tgt Language", style="green")
    table.add_column("Translated Text", style="green")

    table.add_row(text, src_language, tgt_language, translated_text)
    console.print(table)


def print_transliteration_response(
    text: Text, src_language: Text, tgt_language: Text, suggestions: Text
):
    table = Table(show_header=True, header_style="orange3")
    table.add_column("Text", style="sandy_brown")
    table.add_column("Src Language", style="green")
    table.add_column("Tgt Language", style="green")
    table.add_column("Transliterated Suggestions", style="green")

    original_text_shown = False

    for suggestion in suggestions:
        if original_text_shown is False:
            table.add_row(text, src_language, tgt_language, suggestion)
            original_text_shown = True
        else:
            table.add_row("", "", "", suggestion)
    console.print(table)


def print_language_detection_response(text: Text, detected_languages: List[Dict]):
    table = Table(show_header=True, header_style="orange3")
    table.add_column("Text", style="sandy_brown")
    table.add_column("Detected Language", style="green")
    table.add_column("Confidence Score", style="green")
    original_text_shown = False

    for language in detected_languages:
        if original_text_shown is False:
            table.add_row(
                text, language[LANGUAGE], str(round(float(language[CONFIDENCE]), 2))
            )
            original_text_shown = True
        else:
            table.add_row(
                "", language[LANGUAGE], str(round(float(language[CONFIDENCE]), 2))
            )
    console.print(table)


def print_augmentation_response(text: Text, augmentation_suggestions: List):
    table = Table(show_header=True, header_style="orange3")
    table.add_column("Text", style="sandy_brown")
    table.add_column("Suggestions", style="green")
    original_text_shown = False

    for augmentation in augmentation_suggestions:
        if original_text_shown is False:
            table.add_row(text, augmentation)
            original_text_shown = True
        else:
            table.add_row("", augmentation)
    console.print(table)


def print_get_entity(response_to_print):
    table = Table(show_header=True, header_style="orange3")
    table.add_column("idx", style="sandy_brown")
    table.add_column("Entity Name", style="green")
    table.add_column("Description", style="green")
    for index, entity in enumerate(response_to_print[DATA][ENTITIES]):
        table.add_row(str(index), entity[ENTITY], entity[DESCRIPTION])
    console.print(table)


async def get_entity_list(
    language: Text, search: Text, page_number: int, page_size: int, entity_type: Text
):
    payload = {
        FILTER: {ENTITY_TYPE: entity_type, LANGUAGE: language},
        SEARCH: search,
        PAGE_NUMBER: page_number,
        PAGE_SIZE: page_size,
    }
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    async with get_async_http_session().post(
        url=f"{neuralspace_url()}/{LIST_ENTITY_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    ) as response:
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            console.print(
                f"> {INFO} {BOOK} Entities available for Page "
                f"{ORANGE3_START}{page_number}{ORANGE3_END} of {json_response[DATA][COUNT]} "
                f"with Page Size: {ORANGE3_START}{page_size}{ORANGE3_END}"
            )
            print_get_entity(json_response)
        else:
            console.print("Please check your parameter and give the right parameter")



LANGUAGE_CATALOG_URL = "api/nlu/v1/supported-languages"
CREATE_EXAMPLE_URL = "api/nlu/v1/single/example"
CREATE_PROJECT_URL = "api/nlu/v1/project"
LIST_PROJECTS_URL = "api/nlu/v1/list/projects"
LIST_EXAMPLES_URL = "api/nlu/v1/list/example"
DELETE_EXAMPLE_URL = "api/nlu/v1/single/example"
DELETE_PROJECT_URL = "api/nlu/v1/single/project"
TRAIN_MODEL_URL = "api/nlu/v1/model/train/queue"
SINGLE_MODEL_DETAILS_URL = "api/nlu/v1/model"
LIST_MODELS_URL = "api/nlu/v1/list/model"
DEPLOY_MODEL_URL = "api/nlu/v1/model/deploy"
DELETE_MODELS_URL = "api/nlu/v1/model"
PARSE_URL = "api/nlu/v1/model/parse"
CREATE_ENTITY_URL = "api/nlu/v1/entity"
DELETE_ENTITY_URL = "api/nlu/v1/entity"
ADD_ENTITY_EXAMPLES_URL = "api/nlu/v1/add/examples/entity"
DELETE_ENTITY_EXAMPLES_URL = "api/nlu/v1/remove/examples/entity"
DELETE_MULTIPLE_PROJECT_URL = "api/nlu/v1/project"
ADD_SYNONYMS_TO_THE_PROJECT_URL = "api/nlu/v1/project/add/synonym"
LIST_SYNONYMS_IN_THE_PROJECT_URL = "api/nlu/v1/project/list/synonym"


TRAINING_PROGRESS = [
    "Initiated",
    "Queued",
    "Preparing Data",
    "Data Prepared",
    "Pipeline Building",
    "Pipeline Built",
    "Training",
    "Trained",
    "Saved",
    "Completed",
]

SUPPORTED_LANGUAGES = [
    "eu",
    "be",
    "ca",
    "hr",
    "cs",
    "et",
    "gl",
    "hu",
    "ga",
    "la",
    "lv",
    "sr",
    "sk",
    "sl",
    "bg",
    "hy",
    "tr",
    "uk",
    "he",
    "kk",
    "mt",
    "ug",
    "fi",
    "sv",
    "id",
    "ko",
    "vi",
    "af",
    "hi",
    "bn",
    "te",
    "ta",
    "mr",
    "ur",
    "gu",
    "kn",
    "ml",
    "as",
    "pa",
    "fa",
    "ar",
    "el",
    "da",
    "en",
    "nb",
    "zh",
    "nl",
    "fr",
    "de",
    "it",
    "ja",
    "lt",
    "pl",
    "pt",
    "ro",
    "ru",
    "es",
    "sq",
    "an",
    "az",
    "ba",
    "bs",
    "br",
    "my",
    "ce",
    "cv",
    "ka",
    "ht",
    "is",
    "io",
    "jv",
    "ky",
    "lb",
    "mk",
    "mg",
    "ms",
    "ne",
    "oc",
    "su",
    "sw",
    "tl",
    "tg",
    "tt",
    "uz",
    "vo",
    "cy",
    "yo",
    "multilingual",
]

# Training status colour codes
C_INITIATED = "⚪"
C_QUEUED = "🔵"
C_COMPLETED = "🟢"
C_TRAINING = "🟠"
C_FAILED = "🔴"
C_TIMED_OUT = "🟤"
C_DEAD = "⚫"

# DEFAULT VALUES
DEFAULT_RASA_TRAINING_JOBS = 1



import hashlib
import json
import logging
import time
from asyncio import sleep
from copy import copy
from datetime import datetime
from typing import Any, Dict, List, Text, Tuple

import randomname
import requests
from rich.console import Console
from rich.logging import RichHandler
from rich.table import Table

from neuralspace.apis import get_async_http_session
from neuralspace.constants import (
    ARROW_UP,
    AUTHORIZATION,
    BIN,
    BOLD_END,
    BOLD_START,
    BOOK,
    CODE,
    COMMON_HEADERS,
    COMPLETED,
    CONFIDENCE,
    COUNT,
    CREATE_PROJECT_COMMAND,
    CROSS,
    DARK_ORANGE_END,
    DARK_ORANGE_START,
    DATA,
    DEAD,
    DEPLOY_MODEL_COMMAND,
    DEPLOYED,
    DEPLOYED_STATUS,
    DOWN_ARROW,
    END,
    END_INDEX,
    ENTITIES,
    ENTITIES_FILE,
    ENTITY,
    ENTITY_ID,
    ENTITY_PARTIAL_F1,
    ENTITY_TYPE,
    ERROR,
    EXAMPLE,
    EXAMPLE_ID,
    EXAMPLES,
    EXAMPLES_FILE,
    FAILED,
    FAST_FORWARD,
    FILTER,
    FINGER_RIGHT,
    GREEN_TICK,
    HASH,
    INFO,
    INITIATED,
    INTENT,
    INTENT_ACCURACY,
    INTENT_CLASSIFIER_METRICS,
    INTENT_RANKING,
    KEY,
    KEY_DATA,
    KEY_RASA,
    KEY_REPLICAS,
    LANGUAGE,
    LANGUAGES,
    LAST_STATUS_UPDATED,
    LIST_ENTITY_URL,
    LIST_MODELS_COMMAND,
    MESSAGE,
    METRICS,
    MODEL_ID,
    MODEL_NAME,
    MODELS,
    N_REPLICAS,
    NAME,
    NAME_GENERATOR_CONFIG,
    NER_METRICS,
    NO_OF_TRAINING_JOBS,
    NUMBER_OF_EXAMPLES,
    NUMBER_OF_INTENTS,
    NUMBER_OF_MODELS,
    NUMBERS_IN_SQUARE,
    OM,
    ORANGE_END,
    ORANGE_START,
    PAGE_NUMBER,
    PAGE_SIZE,
    PARSE_MODEL_COMMAND,
    PEN_AND_PAPER,
    PERSON_DUMBELL,
    PERSON_HERE,
    PERSON_STANDING,
    PERSON_TAKING,
    PIN,
    PREPARED,
    PROJECT_ID,
    PROJECT_NAME,
    PROJECTS,
    QUEUED,
    RED_END,
    RED_START,
    REPLICAS,
    SAD_SMILEY,
    SAND_CLOCK,
    SAVED,
    SEARCH,
    SOUP,
    START,
    START_INDEX,
    STATUS_MODEL_COMMAND,
    SYNONYMS,
    TEXT,
    TIMED_OUT,
    TRAIN_MODEL_COMMAND,
    TRAINING,
    TRAINING_STATUS,
    TRAINING_TIME,
    TYPE,
    UPLOAD_DATASET_COMMAND,
    USER_DEFINED_ENTITY_TYPES,
    WRITING,
    neuralspace_home,
    neuralspace_url,
    progress_bar,
)
from neuralspace.nlu.constants import (
    ADD_ENTITY_EXAMPLES_URL,
    ADD_SYNONYMS_TO_THE_PROJECT_URL,
    C_COMPLETED,
    C_DEAD,
    C_FAILED,
    C_INITIATED,
    C_QUEUED,
    C_TIMED_OUT,
    C_TRAINING,
    CREATE_ENTITY_URL,
    CREATE_EXAMPLE_URL,
    CREATE_PROJECT_URL,
    DEFAULT_RASA_TRAINING_JOBS,
    DELETE_ENTITY_EXAMPLES_URL,
    DELETE_ENTITY_URL,
    DELETE_EXAMPLE_URL,
    DELETE_MODELS_URL,
    DELETE_MULTIPLE_PROJECT_URL,
    DELETE_PROJECT_URL,
    DEPLOY_MODEL_URL,
    LANGUAGE_CATALOG_URL,
    LIST_EXAMPLES_URL,
    LIST_MODELS_URL,
    LIST_PROJECTS_URL,
    LIST_SYNONYMS_IN_THE_PROJECT_URL,
    PARSE_URL,
    SINGLE_MODEL_DETAILS_URL,
    TRAIN_MODEL_URL,
)
from neuralspace.nlu.utils import print_model_ids
from neuralspace.utils import get_auth_token, is_success_status, print_ner_response

console = Console()

FORMAT = "%(message)s"
logging.basicConfig(
    level="NOTSET", format=FORMAT, datefmt="[%X]", handlers=[RichHandler()]
)

logger = logging.getLogger(__name__)


def create_project_sync(project_name: Text, languages: List[Text]) -> Dict[Text, Any]:
    logger.info(
        f"Creating a project called {project_name} in languages: {', '.join(languages)}"
    )
    payload = {PROJECT_NAME: project_name, LANGUAGE: languages}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    url = f"{neuralspace_url()}/{CREATE_PROJECT_URL}"
    response = requests.post(url, data=json.dumps(payload), headers=HEADERS)
    json_response = response.json()
    if is_success_status(response.status_code):
        logger.info("Successfully created project")
        logger.info(f"Project details: \n {json.dumps(json_response, indent=4)}")
    else:
        logger.error("Failed to create project")
        logger.error(f"Platform response: \n {json.dumps(json_response, indent=4)}")
    return json_response


async def get_languages() -> Dict[Text, Any]:
    console.print(f"> {INFO} {DOWN_ARROW}️ Fetching all supported languages for NLU")
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    async with get_async_http_session().get(
        url=f"{neuralspace_url()}/{LANGUAGE_CATALOG_URL}",
        headers=HEADERS,
    ) as response:
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            console.print(f"> {INFO} {GREEN_TICK} Successfully Fetched Languages")
            table = Table()
            table.add_column("Language")
            table.add_column(CODE)
            for row in json_response[DATA]:
                table.add_row(row[LANGUAGE], row[CODE])
            console.print(table)
            console.print(
                f"{FAST_FORWARD} To create the project: {DARK_ORANGE_START}{CREATE_PROJECT_COMMAND}{DARK_ORANGE_END}"
            )
        else:
            console.print(f"> {ERROR} {CROSS} Failed to create project")
            console.print(
                f'''> Reason for failure {SAD_SMILEY}: " {RED_START}{json_response[MESSAGE]}{RED_END} "'''
            )
    return json_response


async def create_project(project_name: Text, languages: List[Text]) -> Dict[Text, Any]:
    console.print(
        f"> {INFO} Creating a project called "
        f"{project_name} in languages: {', '.join(languages)}!"
    )
    payload = {PROJECT_NAME: project_name, LANGUAGE: languages}
    HEADERS = copy(COMMON_HEADERS)
    table = Table()
    HEADERS[AUTHORIZATION] = get_auth_token()
    async with get_async_http_session().post(
        url=f"{neuralspace_url()}/{CREATE_PROJECT_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    ) as response:
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            console.print(f"> {INFO}" f" {KEY} Retrieving credentials from config...")
            console.print(f"> {INFO} " f"{GREEN_TICK} Successfully created project!")
            console.print(
                f"> {INFO} " f"{PERSON_HERE} Here is your project information..."
            )
            table.add_column("Name")
            table.add_column(LANGUAGE)
            table.add_column("App Type")
            table.add_column("Project Id", style="green")
            language_to_write = ""
            for i, language in enumerate(json_response[DATA][LANGUAGE]):
                if i == len(json_response[DATA][LANGUAGE]) - 1:
                    language_to_write += language
                else:
                    language_to_write += language + ", "
            table.add_row(
                json_response[DATA][PROJECT_NAME],
                language_to_write,
                json_response[DATA]["appType"],
                json_response[DATA][PROJECT_ID],
            )
            console.print(table)
            console.print(
                f"{FAST_FORWARD} Upload data to your project using this command:"
                f" {DARK_ORANGE_START}{UPLOAD_DATASET_COMMAND}{DARK_ORANGE_END}"
            )
        else:
            console.print(f"> {ERROR} {CROSS} Failed to create   project")
            console.print(
                f'''> Reason for failure {SAD_SMILEY}:- " {RED_START}{json_response[DATA]['error']}{RED_END} "'''
            )
    return json_response


async def delete_project(project_id: Text) -> Dict[Text, Any]:
    console.print(f"> {INFO} {BIN} Deleting project with id: {project_id}")
    payload = {PROJECT_ID: project_id}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    async with get_async_http_session().delete(
        url=f"{neuralspace_url()}/{DELETE_PROJECT_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    ) as response:
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            console.print(f"> {INFO} {GREEN_TICK} Successfully deleted project")
            console.print(
                f"{FAST_FORWARD} To create the project:  {DARK_ORANGE_START}{CREATE_PROJECT_COMMAND}{DARK_ORANGE_END}"
            )
        else:
            console.print(f"> {ERROR} {CROSS} Failed to delete projects")
            console.print(
                f"> Reason for failure {SAD_SMILEY}:- {RED_START}{json_response[MESSAGE]}{RED_END}"
            )
    return json_response


def print_projects_table(projects: Dict[Text, Any], verbose: bool):
    table = Table(show_header=True, header_style="bold #c47900", show_lines=True)
    table.add_column("Project Name")
    table.add_column("Project ID")
    if verbose:
        table.add_column("Languages")
        table.add_column("Number of Examples")
        table.add_column("Number of Intents")
        table.add_column("Number of Models")
    for data in projects[DATA][PROJECTS]:
        args = [data[PROJECT_NAME], data[PROJECT_ID]]
        if verbose:
            args += [
                ", ".join(data[LANGUAGE]),
                str(data[NUMBER_OF_EXAMPLES]),
                str(data[NUMBER_OF_INTENTS]),
                str(data[NUMBER_OF_MODELS]),
            ]
        table.add_row(*args)
    console.print(table)


async def list_projects(
    search: Text, page_size: int, page_number: int, languages: List[Text], verbose: bool
) -> Dict[Text, Any]:
    payload = {
        SEARCH: search,
        PAGE_NUMBER: page_number,
        PAGE_SIZE: page_size,
        LANGUAGES: languages,
    }
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    async with get_async_http_session().post(
        url=f"{neuralspace_url()}/{LIST_PROJECTS_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    ) as response:
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            console.print(
                f"> {INFO} {BOOK} Your projects for Page {page_number} "
                f"with Page Size: {page_size}"
            )
            print_projects_table(json_response, verbose)
            console.print(
                f"{FAST_FORWARD} To upload your dataset: {DARK_ORANGE_START}{UPLOAD_DATASET_COMMAND}{DARK_ORANGE_END}"
            )
        else:
            console.print(f"> {ERROR} {CROSS} Failed to list projects")
            console.print(
                f'''> Reason for failure {SAD_SMILEY}: " {RED_START}{json_response[MESSAGE]}{RED_END} "'''
            )
    return json_response


def print_examples_table(examples: Dict[Text, Any], verbose: bool = False):
    table = Table(show_header=True, header_style="bold #c47900", show_lines=True)
    table.add_column("Example ID")
    table.add_column("Text")
    if verbose:
        table.add_column("Intent")
        table.add_column("N Entities", style="#c47900")

    console.print(
        f"> {INFO} {NUMBERS_IN_SQUARE} Total Examples Count: {examples[DATA][COUNT]}"
    )
    for data in examples[DATA][EXAMPLES]:
        marked_sentence = ""
        start_index = []
        end_index = []
        for i, entities in enumerate(data[ENTITIES]):
            start_index.append(entities[START])
            end_index.append(entities[END])
        for idx, character in enumerate(data[TEXT]):
            if idx in start_index:
                marked_sentence += "[bold green]"
            elif idx in end_index:
                marked_sentence += "[/bold green]"
            marked_sentence += character
        args = [data[EXAMPLE_ID], marked_sentence]
        if verbose:
            args += [data[INTENT], str(len(data[ENTITIES]))]
        table.add_row(*args)

    console.print(table)


async def list_examples(
    project_id: Text,
    language: Text,
    prepared: bool,
    type: Text,
    intent: Text,
    page_number: int,
    page_size: int,
    verbose: bool = False,
) -> Dict[Text, Any]:
    console.print(
        f"> {INFO} {DOWN_ARROW}️ Fetching Examples with filter: \n"
        f"> {INFO} {HASH} {BOLD_START}Project ID:{BOLD_END} {ORANGE_START}{project_id}{ORANGE_END}\n"
        f"> {INFO} {OM} Language: {language}\n"
        f"> {INFO} {SOUP} Prepared: {prepared}\n"
        f"> {INFO} {PIN} type: {type}"
    )
    payload = {
        FILTER: {
            PROJECT_ID: project_id,
            LANGUAGE: language,
            PREPARED: prepared,
            TYPE: type,
        },
        PAGE_NUMBER: page_number,
        PAGE_SIZE: page_size,
    }
    if intent:
        payload[FILTER][INTENT] = intent

    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    async with get_async_http_session().post(
        url=f"{neuralspace_url()}/{LIST_EXAMPLES_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    ) as response:
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            print_examples_table(json_response, verbose)
            console.print(
                f"{FAST_FORWARD} To train the model: {DARK_ORANGE_START}{TRAIN_MODEL_COMMAND}{DARK_ORANGE_END}"
            )
            console.print(
                f"{RED_START}{PEN_AND_PAPER}NOTE: To train a model in a project, the project must have minimum"
                f" [orange4]2[/orange4] intent and every intent must atleast have{RED_END} "
                f"[orange4]10[/orange4] {RED_START}training examples{RED_END}"
            )
        else:
            console.print(f"> {ERROR} {CROSS} Failed to list examples")
            console.print(
                f'''> Reason for failure {SAD_SMILEY}: " {RED_START}{json_response[MESSAGE]}{RED_END} "'''
            )

    return json_response


def get_training_status_colour(status: Text) -> Text:
    if status == COMPLETED:
        return C_COMPLETED
    if status == TRAINING:
        return C_TRAINING
    elif status == FAILED:
        return C_FAILED
    elif status == TIMED_OUT:
        return C_TIMED_OUT
    elif status == DEAD:
        return C_DEAD
    elif status == INITIATED:
        return C_INITIATED
    elif status == QUEUED:
        return C_QUEUED
    elif status == SAVED:
        return C_COMPLETED
    elif status == DEPLOYED:
        return C_COMPLETED
    else:
        return C_INITIATED


def print_models_table(models: Dict[Text, Any], verbose: bool):
    table = Table(show_header=True, header_style="bold #c47900", show_lines=True)
    table.add_column("Model ID")
    table.add_column("Model Name")
    if verbose:
        table.add_column("Training Status")
        table.add_column("Replicas")
        table.add_column("Intent Acc")
        table.add_column("Entity Partial F1")
        table.add_column("Training Time (sec)")
        table.add_column("Last Updated")
        table.add_column("Message")

    console.print(
        f"> {INFO} {NUMBERS_IN_SQUARE} Total Models Count: {models[DATA][COUNT]}"
    )
    for data in models[DATA][MODELS]:
        args = [data[MODEL_ID], data[MODEL_NAME]]
        if verbose:
            args += [
                f"{get_training_status_colour(data[TRAINING_STATUS])} {data[TRAINING_STATUS]}",
                str(data[REPLICAS]),
                "{:.3f}".format(
                    data[METRICS][INTENT_CLASSIFIER_METRICS][INTENT_ACCURACY]
                )
                if data[TRAINING_STATUS] == COMPLETED
                else "0.0",
                "{:.3f}".format(data[METRICS][NER_METRICS][ENTITY_PARTIAL_F1])
                if data[TRAINING_STATUS] == COMPLETED
                else "0.0",
                str(data[TRAINING_TIME])
                if data[TRAINING_STATUS] == COMPLETED
                else "0.0",
                str(data[LAST_STATUS_UPDATED]),
                data.get(MESSAGE, ""),
            ]
        table.add_row(*args)
    console.print(table)


async def list_models(
    project_id: Text,
    language: Text,
    training_status: List[Text],
    page_number: int,
    page_size: int,
    verbose: bool,
) -> Dict[Text, Any]:
    console.print(
        f"> {INFO} {DOWN_ARROW}️ Fetching models with filter: \n"
        f"> {INFO} {FINGER_RIGHT} Project ID: {project_id}\n"
        f"> {INFO} {PERSON_TAKING} Language: {language}\n"
        f"> {INFO} {PERSON_DUMBELL}️ Training Statuses: {training_status}"
    )
    payload = {
        FILTER: {PROJECT_ID: project_id, LANGUAGE: language},
        PAGE_NUMBER: page_number,
        PAGE_SIZE: page_size,
    }
    if training_status:
        payload[FILTER][TRAINING_STATUS] = training_status

    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    async with get_async_http_session().post(
        url=f"{neuralspace_url()}/{LIST_MODELS_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    ) as response:
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            print_models_table(json_response, verbose)
            console.print(
                f"{FAST_FORWARD} To deploy the model: {DARK_ORANGE_START}{DEPLOY_MODEL_COMMAND}{DARK_ORANGE_START}"
            )
        else:
            console.print(f"> {ERROR} {CROSS} Failed to list models")
            console.print(
                f'''> Reason for failure {SAD_SMILEY}: " {RED_START}{json_response[MESSAGE]}{RED_END} "'''
            )
    return json_response


async def delete_examples(example_ids: List[Text]) -> Dict[Text, Any]:
    console.print(f"> {INFO} {BIN} Deleting Example with id: {example_ids}")
    for examples_id in example_ids:
        payload = {EXAMPLE_ID: examples_id}
        HEADERS = copy(COMMON_HEADERS)
        HEADERS[AUTHORIZATION] = get_auth_token()
        async with get_async_http_session().delete(
            url=f"{neuralspace_url()}/{DELETE_EXAMPLE_URL}",
            data=json.dumps(payload, ensure_ascii=False),
            headers=HEADERS,
        ) as response:
            json_response = await response.json(encoding="utf-8")
            if is_success_status(response.status):
                console.print(
                    f"> {INFO} {GREEN_TICK} Successfully deleted example_id: {examples_id}"
                )
            else:
                console.print(f"> {ERROR} {CROSS} Failed to delete examples")
                console.print(
                    f'''> Reason for failure {SAD_SMILEY}: " {RED_START}{json_response[MESSAGE]}{RED_END} "'''
                )
    return json_response


async def upload_dataset(
    nlu_data: List[Dict[Text, Text]],
    project_id: Text,
    language: Text,
    skip_first: int = 0,
    ignore_errors: bool = False,
) -> List[Dict[Text, Any]]:
    responses = []
    error_examples = []
    console.print(
        f"> {INFO} Uploading {len(nlu_data) - skip_first} "
        f"examples for project {project_id} and language {language}"
    )
    console.print(f"> {INFO} Skipping first {skip_first} examples")
    Total_number_examples = len(nlu_data[skip_first:])
    current_example_number = 0

    task = progress_bar.add_task(description="", total=Total_number_examples)
    progress_bar.update(task)
    animation_tracker = 1
    for chunk_id, example in enumerate(nlu_data[skip_first:]):
        progress_bar.start()
        current_example_number += 1
        batch = {PROJECT_ID: project_id, LANGUAGE: language, EXAMPLE: example}
        HEADERS = copy(COMMON_HEADERS)
        HEADERS[AUTHORIZATION] = get_auth_token()
        async with get_async_http_session().post(
            url=f"{neuralspace_url()}/{CREATE_EXAMPLE_URL}",
            data=json.dumps(batch, ensure_ascii=False),
            headers=HEADERS,
        ) as response:
            if is_success_status(response.status):
                json_response = await response.json(encoding="utf-8")
                if current_example_number % 3 == 0:
                    if animation_tracker == 1:
                        progress_bar.update(task_id=task, description="🏃")
                        animation_tracker = 2
                    elif animation_tracker == 2:
                        progress_bar.update(task_id=task, description="🚶")
                        animation_tracker = 3
                    elif animation_tracker == 3:
                        progress_bar.update(task_id=task, description="🏃")
                        animation_tracker = 1
                responses.append(json_response)
                progress_bar.update(task_id=task, completed=current_example_number)
            else:
                if "text/html" in response.content_type:
                    message = response.text()
                else:
                    json_response = await response.json(encoding="utf-8")
                    message = json_response[MESSAGE]
                json_response = await response.json(encoding="utf-8")
                console.print(
                    f"> {ERROR} {CROSS} Failed to upload example with text "
                    f"{DARK_ORANGE_START}{example['text']}{DARK_ORANGE_END}"
                )

                console.print(
                    f"> Failed on example: \n{json.dumps(example, indent=4, ensure_ascii=False)}"
                )
                console.print(
                    f'''> Reason for failure {SAD_SMILEY}: " {RED_START}{message}{RED_END} "'''
                )
                error_examples.append(example)
                if ignore_errors:
                    continue
                else:
                    break
    console.print(f"\n> {INFO} {GREEN_TICK} Uploaded {len(responses)} examples")
    console.print(f"> {INFO} {CROSS} Failed on {len(error_examples)} examples")
    with open("failed_examples.json", "w") as f:
        json.dump(error_examples, f, ensure_ascii=False)
        console.print(
            f"> {INFO} {WRITING} Writing failed examples into failed_examples.json"
        )
    console.print(
        f"To train your model: {DARK_ORANGE_START}{TRAIN_MODEL_COMMAND}{DARK_ORANGE_END}"
    )
    return responses


async def wait_till_training_completes(
    model_id: Text, wait: bool, wait_interval: int
) -> Dict[Text, Any]:
    json_response = None
    if wait:
        payload = {
            MODEL_ID: model_id,
        }
        HEADERS = copy(COMMON_HEADERS)
        HEADERS[AUTHORIZATION] = get_auth_token()
        console.print(
            f"> {INFO} {SAND_CLOCK} Waiting for model to "
            f"get trained; model id: {model_id}"
        )
        current_status = ""
        with console.status("...") as status:
            while True:
                async with get_async_http_session().get(
                    url=f"{neuralspace_url()}/{SINGLE_MODEL_DETAILS_URL}",
                    params=payload,
                    headers=HEADERS,
                ) as response:
                    json_response = await response.json(encoding="utf-8")
                    if is_success_status(response.status):
                        current_status = json_response[DATA][TRAINING_STATUS]
                        if (
                            json_response[DATA][TRAINING_STATUS] == COMPLETED
                            or json_response[DATA][TRAINING_STATUS] == FAILED
                            or json_response[DATA][TRAINING_STATUS] == TIMED_OUT
                            or json_response[DATA][TRAINING_STATUS] == DEAD
                        ):
                            if json_response[DATA][TRAINING_STATUS] == COMPLETED:
                                console.print(
                                    f"{FAST_FORWARD} To deploy your model on Neuralspace platform: "
                                    f"{DARK_ORANGE_START}{DEPLOY_MODEL_COMMAND}{DARK_ORANGE_END}"
                                )
                            if json_response[DATA][TRAINING_STATUS] == FAILED:
                                console.print(
                                    f'''> Reason for failure {SAD_SMILEY}: " {RED_START}{json_response[DATA][MESSAGE]}{RED_END}\
                                     "'''
                                )
                            break
                    else:
                        console.print(f"> {ERROR} Failed to fetch model details")
                        console.print(
                            f'''> Reason for failure {SAD_SMILEY}:
                            " {RED_START}{json_response[MESSAGE]}{RED_END} "'''
                        )
                        break
                    status.update(f"Model is {current_status} {PERSON_STANDING}")
                    await sleep(wait_interval)
                    status.update(f"Model is {current_status} {PERSON_DUMBELL}")
            console.print(
                f"> {INFO} "
                f"{get_training_status_colour(json_response[DATA][TRAINING_STATUS])} "
                f"Training status: {json_response[DATA][TRAINING_STATUS]}"
            )
    return json_response


async def train_model(
    project_id: Text, language: Text, model_name: Text, training_jobs: int
) -> Tuple[Dict[Text, Any], Dict[Text, Any]]:
    console.print(
        f"> {INFO} Queuing training job for: \n"
        f"> {INFO} Project ID: {project_id}\n"
        f"> {INFO} Language: {language}\n"
        f"> {INFO} Model Name: {model_name}"
    )
    payload = {
        PROJECT_ID: project_id,
        LANGUAGE: language,
        MODEL_NAME: model_name,
        NO_OF_TRAINING_JOBS: training_jobs,
    }
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()

    async with get_async_http_session().post(
        url=f"{neuralspace_url()}/{TRAIN_MODEL_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    ) as response:
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            model_ids = json_response[DATA]
            console.print(
                f"> {INFO}  Successfully queued {len(model_ids)} training jobs."
            )
            console.print(
                f"To checkout the training status use: {STATUS_MODEL_COMMAND}"
            )
            print_model_ids(model_ids)
            last_model_status = QUEUED
        else:
            console.print(f"> {ERROR} Failed to queue training job")
            console.print(f"> {ERROR} {json_response[MESSAGE]}")
            last_model_status = None
    return json_response, last_model_status


async def delete_models(model_id: Text) -> Dict[Text, Any]:
    console.print(f"> {INFO} {BIN} Deleting model with id: {model_id}")
    payload = {MODEL_ID: model_id}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    async with get_async_http_session().delete(
        url=f"{neuralspace_url()}/{DELETE_MODELS_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    ) as response:
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            console.print(f"> {INFO} {GREEN_TICK} Successfully deleted model")
            console.print(
                f"{FAST_FORWARD} To upload the dataset: {DARK_ORANGE_START}{UPLOAD_DATASET_COMMAND}{DARK_ORANGE_END}"
            )
        else:
            console.print(f"> {ERROR} {CROSS} Failed to delete models")
            console.print(
                f'''> Reason for failure {SAD_SMILEY}: " {RED_START}{json_response[MESSAGE]}{RED_END} "'''
            )
    return json_response


async def deploy(model_id: Text, n_replicas: int) -> Dict[Text, Any]:
    console.print(
        f"> {INFO} {ARROW_UP}️ Deploying: Model ID: {model_id}; Replicas: {n_replicas};"
    )
    payload = {MODEL_ID: model_id, N_REPLICAS: n_replicas}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()

    async with get_async_http_session().post(
        url=f"{neuralspace_url()}/{DEPLOY_MODEL_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    ) as response:
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            console.print(f"> {INFO} {GREEN_TICK} Model added to deployment queue.")
            console.print(
                f"{FAST_FORWARD} Check for replicas in model status to know that the model got deployed "
                f"(use the --verbose flag): "
                f"{DARK_ORANGE_START}{LIST_MODELS_COMMAND}{DARK_ORANGE_END}"
            )
            console.print(
                f"{FAST_FORWARD}Once the model is deployed parse text using: "
                f"{DARK_ORANGE_START}{PARSE_MODEL_COMMAND}{DARK_ORANGE_END}"
            )
        else:
            console.print(f"> {ERROR} {CROSS} Failed to deploy model")
            console.print(
                f'''> Reason for failure {SAD_SMILEY}: " {RED_START}{json_response[MESSAGE]}{RED_END} "'''
            )
    return json_response


def print_nlu_response(nlu_response: Dict[Text, Any], response_time: float):
    table = Table(show_header=True, header_style="bold #c47900", show_lines=True)
    table.add_column("Text")
    table.add_column("Intent")
    table.add_column("Intent Confidence")
    table.add_column("Response Time (m-sec)")
    table.add_row(
        nlu_response[DATA][TEXT],
        nlu_response[DATA][INTENT][NAME],
        str(nlu_response[DATA][INTENT][CONFIDENCE]),
        str(response_time / 1000),
    )
    console.print(table)
    intent_ranking_table = Table(
        show_header=True, header_style="bold #c47900", show_lines=True
    )
    intent_ranking_table.add_column("Intent")
    intent_ranking_table.add_column("Confidence")

    for row in nlu_response[DATA][INTENT_RANKING]:
        intent_ranking_table.add_row(
            row["name"],
            str(row["confidence"]),
        )
    console.print(f"> {INFO} Intent Ranking")
    console.print(intent_ranking_table)
    formatted_entities = []
    for e in nlu_response[DATA][ENTITIES]:
        e[START_INDEX] = e[START]
        e[END_INDEX] = e[END]
        e[TYPE] = e[ENTITY]
        formatted_entities.append(e)
    print_ner_response(formatted_entities, nlu_response[DATA][TEXT])


async def parse(model_id: Text, input_text: Text) -> Dict[Text, Any]:
    console.print(
        f"> {INFO} {PEN_AND_PAPER} Parsing text: {input_text}, using Model ID: {model_id}"
    )
    payload = {MODEL_ID: model_id, TEXT: input_text}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()

    start = datetime.now()
    async with get_async_http_session().post(
        url=f"{neuralspace_url()}/{PARSE_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    ) as response:
        end = datetime.now()
        response_time = (end - start).microseconds
        json_response = await response.json(encoding="utf-8")
        if is_success_status(response.status):
            console.print(f"> {INFO} {GREEN_TICK} Successfully parsed text")
            print_nlu_response(json_response, response_time)
        else:
            console.print(f"> {ERROR} {CROSS} Failed to parse model")
            console.print(
                f'''> Reason for failure {SAD_SMILEY}: " {RED_START}{json_response[MESSAGE]}{RED_END} "'''
            )
    return json_response


def calculate_hash(example: Dict):
    hash = (
        int(
            hashlib.sha256(
                json.dumps(example, ensure_ascii=False).encode("utf-8")
            ).hexdigest(),
            16,
        )
        % 10 ** 8
    )
    return str(hash)


def get_uploaded_examples(project_id) -> Dict[Text, Any]:
    examples_file = neuralspace_home() / KEY_RASA / project_id / EXAMPLES_FILE
    uploaded_examples = {}
    if examples_file.exists():
        with open(examples_file, "r") as exf:
            uploaded_examples = json.load(exf)
    return uploaded_examples


def get_uploaded_entities(project_id) -> Dict[Text, Any]:
    entities_file = neuralspace_home() / KEY_RASA / project_id / ENTITIES_FILE
    uploaded_entities = {}
    if entities_file.exists():
        with open(entities_file) as exf:
            uploaded_entities = json.load(exf)
    return uploaded_entities


def update_examples(uploaded_examples: Dict[Text, Any], project_id: Text):
    examples_file = neuralspace_home() / KEY_RASA / project_id / EXAMPLES_FILE
    if not examples_file.parent.exists():
        examples_file.parent.mkdir(exist_ok=True)
    with open(examples_file, "w") as exf:
        json.dump(uploaded_examples, exf, ensure_ascii=False, indent=4)


def update_entities(uploaded_entities: Dict[Text, Any], project_id: Text):
    examples_file = neuralspace_home() / KEY_RASA / project_id / ENTITIES_FILE
    if not examples_file.parent.exists():
        examples_file.parent.mkdir(exist_ok=True)
    with open(examples_file, "w") as exf:
        json.dump(uploaded_entities, exf, indent=4)


def upload_example_sync(
    example: Dict[Text, Any], project_id: Text, language: Text
) -> requests.Response:
    logger.debug(f"Uploading example: {example}")
    batch = {PROJECT_ID: project_id, LANGUAGE: language, EXAMPLE: example}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()

    url = f"{neuralspace_url()}/{CREATE_EXAMPLE_URL}"
    response = requests.post(
        url, data=json.dumps(batch, ensure_ascii=False), headers=HEADERS
    )
    return response


def upload_entity_sync(example: Dict[Text, Any]) -> requests.Response:
    console.print(f"Uploading entity: {example}")
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    url = f"{neuralspace_url()}/{CREATE_ENTITY_URL}"
    response = requests.post(
        url, data=json.dumps(example, ensure_ascii=False), headers=HEADERS
    )
    return response


def start_training(project_id: Text, language: Text) -> Dict[Text, Any]:
    data = {
        PROJECT_ID: project_id,
        LANGUAGE: language,
        NO_OF_TRAINING_JOBS: DEFAULT_RASA_TRAINING_JOBS,
    }
    url = f"{neuralspace_url()}/{TRAIN_MODEL_URL}"
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    response = requests.post(url=url, json=data, headers=HEADERS)
    json_response: Dict = response.json(encoding="utf-8")
    if not is_success_status(response.status_code):
        raise Exception(json_response.get(MESSAGE))
    return json_response


def wait_till_training_completes_sync(
    model_id: Text,
    wait_interval: int,
) -> Dict[Text, Any]:
    payload = {
        MODEL_ID: model_id,
    }
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    console.print(
        f"> {INFO} {SAND_CLOCK} Waiting for model to "
        f"get trained; model id: {model_id}"
    )
    with console.status("...") as status:
        while True:
            response = requests.get(
                url=f"{neuralspace_url()}/{SINGLE_MODEL_DETAILS_URL}",
                params=payload,
                headers=HEADERS,
            )
            json_response: Dict[Text, Any] = response.json(encoding="utf-8")
            if is_success_status(response.status_code):
                current_status = json_response[DATA][TRAINING_STATUS]
                if (
                    json_response[DATA][TRAINING_STATUS] == COMPLETED
                    or json_response[DATA][TRAINING_STATUS] == FAILED
                    or json_response[DATA][TRAINING_STATUS] == TIMED_OUT
                    or json_response[DATA][TRAINING_STATUS] == DEAD
                ):
                    if json_response[DATA][TRAINING_STATUS] == COMPLETED:
                        pass
                    if json_response[DATA][TRAINING_STATUS] == FAILED:
                        console.print(
                            f'''> Reason for failure {SAD_SMILEY}: " {RED_START}{json_response[DATA]['message']}{RED_END}\
                                                 "'''
                        )
                        raise Exception("Model training failed")
                    break
            else:
                console.print(f"> {ERROR} Failed to fetch model details")
                console.print(
                    f'''> Reason for failure {SAD_SMILEY}:
                                        " {RED_START}{json_response['message']}{RED_END} "'''
                )
                raise Exception("Model training failed")
            status.update(f"Model is {current_status} {PERSON_STANDING}")
            time.sleep(wait_interval)
            status.update(f"Model is {current_status} {PERSON_DUMBELL}")
        console.print(
            f"> {INFO} "
            f"{get_training_status_colour(json_response[DATA][TRAINING_STATUS])} "
            f"Training status: {json_response[DATA][TRAINING_STATUS]}"
        )
    return json_response


def wait_till_deploy_completes_sync(
    model_id: Text,
    wait_interval: int,
    timeout_in_mins: int = 5,
    check_n_replicas: int = 1,
) -> Dict[Text, Any]:
    payload = {
        MODEL_ID: model_id,
    }
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    start_time = datetime.now()
    with console.status("...") as status:
        while True:
            response = requests.get(
                url=f"{neuralspace_url()}/{SINGLE_MODEL_DETAILS_URL}",
                params=payload,
                headers=HEADERS,
            )
            json_response: Dict[Text, Any] = response.json(encoding="utf-8")
            current_status = json_response[DATA][DEPLOYED_STATUS]
            if is_success_status(response.status_code):
                if (
                    json_response[DATA][KEY_REPLICAS] == check_n_replicas
                    and current_status == DEPLOYED
                ):

                    break
                elif json_response[DATA][DEPLOYED_STATUS] == FAILED:
                    console.print(
                        f'''> Reason for failure {SAD_SMILEY}: " {RED_START}{json_response[DATA]['message']}{RED_END}\
                                             "'''
                    )
                    raise Exception(
                        f"Model deployment failed because: {json.dumps(response.json(), indent=4)}"
                    )
            else:
                console.print(f"> {ERROR} Failed to fetch model details")
                console.print(
                    f'''> Reason for failure {SAD_SMILEY}:
                                        " {RED_START}{json_response[MESSAGE]}{RED_END} "'''
                )
                raise Exception("Model training failed")
            status.update(
                f"Model is {current_status} with currently "
                f"{json_response[DATA][KEY_REPLICAS]} replicas {PERSON_STANDING}"
            )
            time.sleep(wait_interval)
            status.update(
                f"Model is {current_status} with currently "
                f"{json_response[DATA][KEY_REPLICAS]} replicas {PERSON_DUMBELL}"
            )
            end_time = datetime.now()
            time_diff_in_mins = (end_time - start_time).seconds / 60
            if time_diff_in_mins >= timeout_in_mins:
                raise Exception(
                    f"Timeout while waiting for model to get deployed. Timeout set to {timeout_in_mins}"
                )
        if check_n_replicas != 0:
            console.print(
                f"> {INFO} "
                f"{get_training_status_colour(current_status)} "
                f"Deployment status: {current_status}"
            )
    return json_response


def parse_sync(model_id: Text, input_text: Text) -> Dict[Text, Any]:
    logger.info(f"Parsing text: {input_text}, using Model ID: {model_id}")
    payload = {MODEL_ID: model_id, TEXT: input_text}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    print(json.dumps(payload))
    start = datetime.now()
    response = requests.post(
        url=f"{neuralspace_url()}/{PARSE_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    )
    print(f"Response is :: {response.text}")
    end = datetime.now()
    response_time = (end - start).microseconds
    json_response = response.json(encoding="utf-8")
    if is_success_status(response.status_code):
        logger.debug(
            f"Platform Response: \n{json.dumps(json_response, indent=4, ensure_ascii=False)}"
        )
        print_nlu_response(json_response, response_time)
    else:
        logger.error("Failed to deploy model")
        logger.error(f"Platform response: \n {json.dumps(json_response, indent=4)}")
    return json_response


def deploy_sync(model_id: Text, n_replicas: int) -> Dict[Text, Any]:
    if n_replicas == 0:
        console.print(f"> {INFO} Un-deploying the Previous model: {model_id}")
    else:
        console.print(f"> {INFO} Deploying the current model: {model_id}")
    payload = {MODEL_ID: model_id, N_REPLICAS: n_replicas}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    response = requests.post(
        url=f"{neuralspace_url()}/{DEPLOY_MODEL_URL}", json=payload, headers=HEADERS
    )
    json_response = response.json(encoding="utf-8")
    return json_response


def delete_examples_sync(example_id: Text) -> Dict[Text, Any]:
    console.print(f"> {INFO} {BIN} Deleting Example with id: {example_id}")
    payload = {EXAMPLE_ID: example_id}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    response = requests.delete(
        url=f"{neuralspace_url()}/{DELETE_EXAMPLE_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    )
    json_response = response.json(encoding="utf-8")
    if is_success_status(response.status_code):
        logger.debug(
            f"> {INFO} {GREEN_TICK} Successfully deleted example: {example_id}!"
        )
    else:
        raise Exception(
            f"> {ERROR} {CROSS} Failed to add entity examples {SAD_SMILEY}: "
            f"{RED_START}{json_response['message']}{RED_END}"
        )
    return json_response


def delete_entity_sync(entity_id: Text) -> Dict[Text, Any]:
    console.print(f"> {INFO} {BIN} Deleting entity with id: {entity_id}")
    payload = {ENTITY_ID: entity_id}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    response = requests.delete(
        url=f"{neuralspace_url()}/{DELETE_ENTITY_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    )
    json_response = response.json(encoding="utf-8")
    if is_success_status(response.status_code):
        logger.debug(f"> {INFO} {GREEN_TICK} Successfully deleted entity!")
    else:
        raise Exception(
            f'''> Failed to delete entity {SAD_SMILEY}: " {json_response['message']}"'''
        )
    return json_response


def delete_entity_examples_sync(
    entity_id: Text, examples: List[Text]
) -> Dict[Text, Any]:
    logger.debug(f"> {INFO} {BIN} Deleting entity with id: {entity_id}")
    payload = {ENTITY_ID: entity_id, EXAMPLES: examples}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    response = requests.post(
        url=f"{neuralspace_url()}/{DELETE_ENTITY_EXAMPLES_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    )
    json_response = response.json(encoding="utf-8")
    if is_success_status(response.status_code):
        logger.debug(f"> {INFO} {GREEN_TICK} Successfully deleted entity examples!")
    else:
        raise Exception(
            f"{ERROR} {CROSS} Failed to delete entity examples {SAD_SMILEY}: "
            f"{RED_START}{json_response['message']}{RED_END} "
        )

    return json_response


def add_entity_examples_sync(entity_id: Text, examples: List[Text]) -> Dict[Text, Any]:
    logger.debug(
        f"> {INFO} {BIN} Adding examples {examples} to entity with id: {entity_id}"
    )
    payload = {ENTITY_ID: entity_id, EXAMPLES: examples}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    response = requests.post(
        url=f"{neuralspace_url()}/{ADD_ENTITY_EXAMPLES_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    )
    json_response = response.json(encoding="utf-8")
    if is_success_status(response.status_code):
        logger.debug(f"> {INFO} {GREEN_TICK} Successfully added entity examples!")
    else:
        raise Exception(
            f"> {ERROR} {CROSS} Failed to add entity examples {SAD_SMILEY}:  "
            f"{RED_START}{json_response['message']}{RED_END}"
        )
    return json_response


def create_project_id(language: Text):
    console.print(f"Creating a new project with language {language}")
    result = create_project_sync(
        project_name=f"rasa-{randomname.get_name(**NAME_GENERATOR_CONFIG)}-{language}-{datetime.now()}",
        languages=[language],
    )
    return result[KEY_DATA][PROJECT_ID]


def list_projects_sync(
    search: Text, page_size: int, page_number: int, languages: List[Text], verbose: bool
) -> Dict[Text, Any]:
    payload = {
        SEARCH: search,
        PAGE_NUMBER: page_number,
        PAGE_SIZE: page_size,
        LANGUAGES: languages,
    }
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    response = requests.post(
        url=f"{neuralspace_url()}/{LIST_PROJECTS_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    )
    json_response = response.json(encoding="utf-8")
    return json_response


def delete_multiple_project_sync(project_ids: List[Text]):
    payload = {PROJECT_ID: project_ids}
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    response = requests.delete(
        url=f"{neuralspace_url()}/{DELETE_MULTIPLE_PROJECT_URL}",
        data=json.dumps(payload, ensure_ascii=False),
        headers=HEADERS,
    )
    json_response = response.json(encoding="utf-8")
    return json_response


def delete_all_project(total_projects: int):
    project_ids = []
    if not total_projects == 0:
        all_projects = list_projects_sync(
            "", page_size=total_projects, page_number=1, languages=[], verbose=False
        )
        all_projects = all_projects[DATA][PROJECTS]
        for project in all_projects:
            project_ids.append(project[PROJECT_ID])
        delete_multiple_project_sync(project_ids)


def list_entity_sync(
    entity_type: Text, page_size: int, page_number: int, search: Text, language: Text
) -> Dict[Text, Any]:
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    payload = {
        FILTER: {ENTITY_TYPE: entity_type, LANGUAGE: language},
        SEARCH: search,
        PAGE_NUMBER: page_number,
        PAGE_SIZE: page_size,
    }
    response = requests.post(
        url=f"{neuralspace_url()}/{LIST_ENTITY_URL}",
        headers=HEADERS,
        data=json.dumps(payload),
    )
    if not is_success_status(response.status_code):
        raise Exception(
            "Error while Getting the entities from the platform please try again"
            "after some time or raise a query in the slack channel"
        )

    json_response = response.json()
    return json_response


def list_all_entity_sync(language: Text) -> Dict[Text, Dict[Text, Any]]:
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    entities_names = {}
    for entity_type in USER_DEFINED_ENTITY_TYPES:
        json_response = list_entity_sync(
            entity_type, page_size=1, page_number=1, search="", language=language
        )
        logger.debug(json_response)
        count = json_response[DATA][COUNT]
        if not count == 0:
            json_response = list_entity_sync(
                entity_type,
                page_size=count,
                page_number=1,
                search="",
                language=language,
            )
            entities_names[entity_type] = {
                entity[ENTITY]: {
                    ENTITY_ID: entity[ENTITY_ID],
                    EXAMPLES: entity[EXAMPLES],
                }
                for entity in json_response[DATA][ENTITIES]
            }
        else:
            entities_names[entity_type] = {}
    return entities_names


def add_synonym_to_project_sync(project_id: Text, language: Text, entity_id: Text):
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    payload = {
        PROJECT_ID: project_id,
        LANGUAGE: language,
        ENTITY_ID: entity_id,
    }
    response = requests.post(
        url=f"{neuralspace_url()}/{ADD_SYNONYMS_TO_THE_PROJECT_URL}",
        headers=HEADERS,
        data=json.dumps(payload),
    )
    if not is_success_status(response.status_code):
        raise Exception(
            "Error while Adding the synonym from the platform please try again"
            "after some time or raise a query in the slack channel"
        )

    json_response = response.json()
    return json_response


def list_project_synonym_sync(
    project_id: Text, language: Text, search: Text, page_number: int, page_size: int
):
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    payload = {
        PROJECT_ID: project_id,
        LANGUAGE: language,
        SEARCH: search,
        PAGE_NUMBER: page_number,
        PAGE_SIZE: page_size,
    }
    response = requests.post(
        url=f"{neuralspace_url()}/{LIST_SYNONYMS_IN_THE_PROJECT_URL}",
        headers=HEADERS,
        data=json.dumps(payload),
    )
    if not is_success_status(response.status_code):
        raise Exception(
            "Error while Getting the project synonyms from the platform please try again"
            "after some time or raise a query in the slack channel"
        )

    json_response = response.json()
    return json_response


def list_all_project_synonym_sync(project_id: Text, language: Text):
    HEADERS = copy(COMMON_HEADERS)
    HEADERS[AUTHORIZATION] = get_auth_token()
    synonym_list = {}
    json_response = list_project_synonym_sync(
        project_id, page_size=1, page_number=1, search="", language=language
    )
    logger.debug(json_response)
    count = json_response[DATA][COUNT]
    if not count == 0:
        json_response = list_project_synonym_sync(
            project_id,
            page_size=count,
            page_number=1,
            search="",
            language=language,
        )
        synonym_list = {
            synonym[ENTITY]: synonym[ENTITY_ID]
            for synonym in json_response[DATA][SYNONYMS]
        }
    return synonym_list
